# -*- coding: utf-8 -*-
"""
    weppy.cache
    -----------

    Provides a caching system.

    :copyright: (c) 2014-2016 by Giovanni Barillari
    :license: BSD, see LICENSE for more details.
"""

import os
import time
import heapq
import threading
import tempfile

from ._compat import pickle, integer_types, iteritems, hashlib_md5
from .libs.portalocker import LockedFile

__all__ = ['Cache']


class BaseCache(object):
    def __init__(self, default_expire=300):
        self._default_expire = default_expire

    def __call__(self, key, function=None, dt='default'):
        return self._get(key)

    def _get(self, key):
        return None

    def _set(self, key, value, dt):
        return None

    def clear(self, key=None):
        return None

    #def inc(self, key, dt=1):
    #    return
    #
    #def dec(self, key, dt=1):
    #    return


class RamElement(object):
    def __init__(self, value, exp, acc):
        self.value = value
        self.exp = exp
        self.acc = acc


class RamCache(BaseCache):
    lock = threading.RLock()
    map = {}
    _heap_exp = []
    _heap_acc = []

    def __init__(self, prefix='', threshold=500, default_expire=300):
        BaseCache.__init__(self, default_expire)
        self._prefix = prefix
        self._threshold = threshold

    def __call__(self, key, function=None, dt='default'):
        key = self._prefix + key
        now = time.time()
        if dt is None:
            dt = 60 * 60 * 24 * 365
        if dt == "default":
            dt = self._default_expire
        if function:
            value = self._get(key)
            if value is None:
                value = function()
                expiration = now + dt
                self._set(key, value, expiration)
            return value
        return None

    def clear(self, key=None):
        with self.lock:
            if key is not None:
                key = self._prefix + key
                try:
                    rv = self.map[key]
                    self._heap_acc.remove((rv.acc, key))
                    self._heap_exp.remove((rv.exp, key))
                    del self.map[key]
                    return
                except:
                    return
            self.map.clear()
            self._heap_acc = []
            self._heap_exp = []

    def _prune(self):
        now = time.time()
        # remove expired items
        while self._heap_exp:
            exp, rk = heapq.heappop(self._heap_exp)
            if exp < now:
                self._heap_acc.remove((self.map[rk].acc, rk))
                del self.map[rk]
            else:
                heapq.heappush(self._heap_exp, (exp, rk))
                break
        # remove threshold exceding elements
        while len(self.map) > self._threshold:
            rk = heapq.heappop(self._heap_acc)[1]
            self._heap_exp.remove((self.map[rk].exp, rk))
            del self.map[rk]

    def _get(self, key):
        try:
            with self.lock:
                element = self.map[key]
                now = time.time()
                if element.exp < now:
                    return None
                self._heap_acc.remove((element.acc, key))
                element.acc = now
                heapq.heappush(self._heap_acc, (element.acc, key))
            val = element.value
        except KeyError:
            return None
        return val

    def _set(self, key, value, dt):
        with self.lock:
            self._prune()
            now = time.time()
            heapq.heappush(self._heap_exp, (dt, key))
            heapq.heappush(self._heap_acc, (now, key))
            self.map[key] = RamElement(value, dt, now)


class DiskCache(BaseCache):
    lock = threading.RLock()
    _fs_transaction_suffix = '.__wp_cache'
    _fs_mode = 0o600

    def __init__(self, cache_dir='cache', threshold=500, default_expire=300):
        BaseCache.__init__(self, default_expire)
        self._threshold = threshold
        from .expose import Expose
        self._path = os.path.join(Expose.application.root_path, cache_dir)
        #: create required paths if needed
        if not os.path.exists(self._path):
            os.mkdir(self._path)

    def __call__(self, key, function=None, dt='default'):
        now = time.time()
        if dt is None:
            dt = 60 * 60 * 24 * 365
        if dt == "default":
            dt = self._default_expire
        if function:
            value = self._get(key)
            if value is None:
                value = function()
                expiration = now + dt
                self._set(key, value, expiration)
            return value
        return None

    def clear(self, key=None):
        with self.lock:
            if key is not None:
                filename = self._get_filename(key)
                try:
                    os.remove(filename)
                    return
                except:
                    return
            for name in self._list_dir():
                self._del_file(name)

    def _list_dir(self):
        return [os.path.join(self._path, fn) for fn in os.listdir(self._path)
                if not fn.endswith(self._fs_transaction_suffix)]

    def _prune(self):
        with self.lock:
            entries = self._list_dir()
            if len(entries) > self._threshold:
                now = time.time()
                try:
                    for i, fpath in enumerate(entries):
                        remove = False
                        f = LockedFile(fpath, 'rb')
                        #with open(fpath, 'rb') as f:
                        exp = pickle.load(f.file)
                        f.close()
                        remove = exp <= now or i % 3 == 0
                        if remove:
                            self._del_file(fpath)
                except:
                    pass

    def _get_filename(self, key):
        khash = hashlib_md5(key).hexdigest()
        return os.path.join(self._path, khash)

    def _del_file(self, filename):
        try:
            os.remove(filename)
        except:
            pass

    def _get(self, key):
        filename = self._get_filename(key)
        try:
            with self.lock:
                now = time.time()
                f = LockedFile(filename, 'rb')
                #with open(filename, 'rb') as f:
                exp = pickle.load(f.file)
                if exp < now:
                    f.close()
                    return None
                val = pickle.load(f.file)
                f.close()
        except:
            return None
        return val

    def _set(self, key, value, exp):
        filename = self._get_filename(key)
        with self.lock:
            self._prune()
            try:
                fd, tmp = tempfile.mkstemp(suffix=self._fs_transaction_suffix,
                                           dir=self._path)
                with os.fdopen(fd, 'wb') as f:
                    pickle.dump(exp, f, 1)
                    pickle.dump(value, f, pickle.HIGHEST_PROTOCOL)
                os.rename(tmp, filename)
                os.chmod(filename, self._fs_mode)
            except:
                pass


class RedisCache(BaseCache):
    def __init__(self, host='localhost', port=6379, password=None,
                 db=0, default_expire=300, prefix='', **kwargs):
        BaseCache.__init__(self, default_expire=default_expire)
        try:
            import redis
        except ImportError:
            raise RuntimeError('no redis module found')
        self._cache = redis.Redis(host=host, port=port, password=password,
                                  db=db, **kwargs)
        self._prefix = prefix

    def __call__(self, key, function=None, dt='default'):
        key = self._prefix + key
        if dt is None:
            dt = 60 * 60 * 24 * 365
        if dt == "default":
            dt = self._default_expire
        if function:
            value = self._get(key)
            if value is None:
                value = function()
                self._set(key, value, dt)
            return value
        return None

    def clear(self, key=None):
        if key is not None:
            key = self._prefix + key
            self._cache.delete(key)
            return
        if self._prefix:
            keys = self._cache.keys(self._prefix + '*')
            if keys:
                self._cache.delete(*keys)
        else:
            self._cache.flushdb()

    def _dump_obj(self, value):
        if type(value) in integer_types:
            return str(value).encode('ascii')
        return b'!' + pickle.dumps(value)

    def _load_obj(self, value):
        if value is None:
            return None
        if value.startswith(b'!'):
            try:
                return pickle.loads(value[1:])
            except pickle.PickleError:
                return None
        try:
            return int(value)
        except ValueError:
            return None

    def _get(self, key):
        return self._load_obj(self._cache.get(key))

    def _set(self, key, value, dt):
        dumped = self._dump_obj(value)
        return self._cache.setex(name=key, value=dumped, time=dt)


class Cache(object):
    def __init__(self, **kwargs):
        #: load handlers
        handlers = []
        for key, val in iteritems(kwargs):
            if key == "default":
                continue
            handlers.append((key, val))
        if not handlers:
            handlers.append(('ram', RamCache()))
        #: set handlers
        for name, handler in handlers:
            setattr(self, name, handler)
        self.default_handler = kwargs.get('default', handlers[0][0])

    def __call__(self, key, f, time_expire=None):
        return self.__getattribute__(self.default_handler)(key, f, time_expire)



from .adapters import adapters as adapters_registry
from .base import DAL, Field
from .models import Model
from .apis import belongs_to, refers_to, has_one, has_many, compute, \
    rowattr, rowmethod, before_insert, before_update, before_delete, \
    after_insert, after_update, after_delete, scope



#!/usr/bin/env python
# -*- coding: utf8 -*-
# Plural-Forms for uk (Ukrainian)

# Ukrainian language has 3 forms:  1 singular and 2 plurals
nplurals = 3


# Determine plural_id for number *n* as sequence of positive
# integers: 0,1,...
# NOTE! For singular form ALWAYS return plural_id = 0
get_plural_id = lambda n: (0 if n % 10 == 1 and n % 100 != 11 else
                           1 if n % 10 >= 2 and n % 10 <= 4 and
                           (n % 100 < 10 or n % 100 >= 20) else
                           2)

# construct_plural_form() is not used now because of complex
# rules of Ukrainian language. Default version of
# this function is used to simple insert new words into
# plural_dict dictionary)
# construct_plural_form = lambda word, plural_id: word



# -*- coding: utf-8 -*-
"""
    weppy.tools.auth.exposer
    ------------------------

    Provides the exposed functions and logics for the authorization system.

    :copyright: (c) 2014-2016 by Giovanni Barillari

    Based on the web2py's auth module (http://www.web2py.com)
    :copyright: (c) by Massimo Di Pierro <mdipierro@cs.depaul.edu>

    :license: LGPLv3 (http://www.gnu.org/licenses/lgpl.html)
"""

import time
from ...dal import Field
from ...forms import Form, DALForm
from ...globals import request, session
from ...helpers import flash, abort
from ...http import redirect
from .handlers import DefaultLoginHandler
from .helpers import callback, replace_id, get_vars_next


class Exposer(object):
    def __init__(self, auth):
        self.auth = auth
        self.settings = auth.settings
        self.messages = auth.messages
        self.form_data = {}
        self._build_register_form_()
        self._build_retrieve_username_form_()
        self._build_reset_password_form_()
        self._build_request_reset_password_form_()
        self._build_change_password_form_()

    def _build_register_form_(self):
        if not self.settings.register_fields:
            self.settings.register_fields = {'writable': [
                field.name for field in self.auth.table_user
                if field.type != 'id' and field.writable
            ]}
        all_fieldkeys = [
            field.name for field in self.auth.table_user if field.name
            in self.settings.register_fields['writable']
        ]
        for i, fieldname in enumerate(all_fieldkeys):
            if fieldname == 'password':
                all_fieldkeys.insert(i + 1, 'password2')
                break
        form_fields = {}
        for i, fieldname in enumerate(all_fieldkeys):
            if fieldname != 'password2':
                form_fields[fieldname] = \
                    self.auth.table_user[fieldname].clone()
            else:
                form_fields[fieldname] = Field(
                    'password', label=self.messages.verify_password
                )
            form_fields[fieldname]._inst_count_ = i
        self.form_data['register'] = form_fields

    def _build_retrieve_username_form_(self):
        form_fields = {
            'email': Field(
                validation={'is': 'email', 'presence': True},
                label='E-mail'
            )
        }
        self.form_data['retrieve_username'] = form_fields

    def _build_reset_password_form_(self):
        form_fields = {
            'password': self.auth.table_user['password'].clone(),
            'password2': Field(
                'password', label=self.messages.verify_password
            )
        }
        self.form_data['reset_password'] = form_fields

    def _build_request_reset_password_form_(self):
        userfield = self.settings.login_userfield
        if userfield == 'email':
            loginfield = Field(
                validation={'is': 'email', 'presence': True}
            )
        else:
            v = {'presence': True}
            if self.settings.username_case_sensitive:
                v['lower'] = True
            loginfield = Field(validation=v)
        form_fields = {
            userfield: loginfield
        }
        self.form_data['request_reset_password'] = form_fields

    def _build_change_password_form_(self):
        form_fields = {
            'old_password': self.auth.table_user['password'].clone(),
            'new_password': self.auth.table_user['password'].clone(),
            'new_password2': Field(
                'password', label=self.messages.verify_password
            )
        }
        form_fields['old_password'].label = self.messages.old_password
        form_fields['new_password'].lable = self.messages.new_password
        self.form_data['change_password'] = form_fields

    def login(self):
        return self.auth._login_with_handler(DefaultLoginHandler)

    def logout(self):
        nextv = (get_vars_next() or self.settings.logout_next or
                 self.auth.url('login'))
        onlogout = self.settings.logout_onlogout
        if onlogout:
            onlogout(self.auth.user)
        log = self.messages['logout_log']
        if self.auth.user:
            self.auth.log_event(log, self.auth.user)
        if self.settings.login_form != self.auth:
            cas = self.settings.login_form
            cas_user = cas.get_user()
            if cas_user:
                nextv = cas.logout_url(nextv)
        session.auth = None
        flash(self.messages.logged_out)
        if nextv is not None:
            redirect(nextv)

    def register(self):
        def process_form(form):
            if form.params.password.password != form.params.password2.password:
                form.errors.password = "password mismatch"
                form.errors.password2 = "password mismatch"
                return
            for validation in onvalidation:
                validation(form)

        if self.auth.is_logged_in():
            redirect(self.settings.url_logged or self.auth.url('profile'))
        nextv = get_vars_next() or self.settings.register_next
        onvalidation = self.settings.register_onvalidation
        onaccept = self.settings.register_onaccept
        log = self.messages['register_log']
        username = self.settings.login_userfield
        form = Form(
            self.form_data['register'],
            hidden=dict(_next=nextv),
            submit=self.messages.register_button,
            onvalidation=process_form,
            keepvalues=True
        )
        if form.accepted:
            del form.params['password2']
            # insert user
            form.params.id = self.auth.table_user.insert(**form.params)
            row = self.auth.table_user(id=form.params.id)
            description = self.messages.group_description % form.params
            if self.settings.create_user_groups:
                group_id = self.auth.add_group(
                    self.settings.create_user_groups % form.params,
                    description)
                self.add_membership(group_id, form.params.id)
            if self.settings.everybody_group_id:
                self.auth.add_membership(
                    self.settings.everybody_group_id, form.params.id)
            if self.settings.registration_requires_verification:
                link = self.auth.url(
                    ['verify_email', row['registration_key']], scheme=True
                )
                d = dict(request.params)
                d.update(dict(key=row['registration_key'], link=link,
                         username=form.params[username]))
                if not (self.settings.mailer and self.settings.mailer.send(
                        to=form.params.email,
                        subject=self.messages.verify_email_subject,
                        message=self.messages.verify_email % d)):
                    self.auth.db.rollback()
                    flash(self.messages.unable_send_email)
                    return form
                flash(self.messages.email_sent)
            if self.settings.registration_requires_approval and \
               not self.settings.registration_requires_verification:
                row.update_record(registration_key='pending')
                flash(self.messages.registration_pending)
            elif (not self.settings.registration_requires_verification or
                    self.settings.login_after_registration):
                flash(self.messages.registration_successful)
                self.auth.login_user(row)
                flash(self.messages.logged_in)
            self.auth.log_event(log, form.params)
            callback(onaccept, form)
            if not nextv:
                nextv = self.auth.url('login')
            else:
                nextv = replace_id(nextv, form)
            redirect(nextv)
        return form

    def verify_email(self, key):
        user = self.auth.table_user(registration_key=key)
        if not user:
            redirect(self.settings.login_url or self.auth.url('login'))
        if self.settings.registration_requires_approval:
            user.update_record(registration_key='pending')
            flash(self.messages.registration_pending)
        else:
            user.update_record(registration_key='')
            flash(self.messages.email_verified)
        # make sure session has same user.registration_key as db record
        if self.auth.user:
            self.auth.user.registration_key = user.registration_key
        log = self.messages['verify_email_log']
        nextv = self.settings.verify_email_next or self.auth.url('login')
        onaccept = self.settings.verify_email_onaccept
        self.auth.log_event(log, user)
        callback(onaccept, user)
        redirect(nextv)

    def retrieve_username(self):
        if 'username' not in self.auth.table_user.fields:
            raise abort(404)
        if not self.settings.mailer:
            flash(self.messages.function_disabled)
            return ''
        nextv = get_vars_next() or self.settings.retrieve_username_next
        onvalidation = self.settings.retrieve_username_onvalidation
        onaccept = self.settings.retrieve_username_onaccept
        log = self.messages['retrieve_username_log']
        form = Form(
            self.form_data['retrieve_username'],
            hidden=dict(_next=nextv),
            sumbit=self.messages.submit_button,
            onvalidation=onvalidation
        )
        if form.accepted:
            users = self.auth.db(
                self.auth.table_user.email == form.params.email).select()
            if not users:
                flash(self.messages.invalid_email)
                redirect(self.auth.url('retrieve_username'))
            username = ', '.join(u.username for u in users)
            self.settings.mailer.send(
                to=form.params.email,
                subject=self.messages.retrieve_username_subject,
                message=self.messages.retrieve_username % dict(
                    username=username))
            flash(self.messages.email_sent)
            for user in users:
                self.auth.log_event(log, user)
            callback(onaccept, form)
            if not nextv:
                nextv = self.auth.url('retrieve_username')
            else:
                nextv = replace_id(nextv, form)
            redirect(nextv)
        return form

    def reset_password(self):
        def process_form(form):
            if form.params.password.password != form.params.password2.password:
                form.errors.password = self.messages.mismatched_password
                form.errors.password2 = self.messages.mismatched_password

        nextv = get_vars_next() or self.settings.reset_password_next
        try:
            key = request.params.key
            t0 = int(key.split('-')[0])
            if time.time() - t0 > 60 * 60 * 24:
                raise Exception
            user = self.auth.table_user(reset_password_key=key)
            if not user:
                raise Exception
        except Exception:
            flash(self.messages.invalid_reset_password)
            redirect(nextv)
        form = Form(
            self.form_data['reset_password'],
            onvalidation=process_form,
            submit=self.messages.password_reset_button,
            hidden=dict(_next=nextv),
        )
        if form.accepted:
            user.update_record(
                password=str(form.params.new_password),
                registration_key='',
                reset_password_key=''
            )
            flash(self.messages.password_changed)
            if self.settings.login_after_password_change:
                self.auth.login_user(user)
            redirect(nextv)
        return form

    def request_reset_password(self):
        def process_form(form, rows):
            field = self.settings.login_userfield
            user = self.auth.table_user(**{field: form.params.email})
            rows['user'] = user
            if not user:
                form.errors[field] = self.messages['invalid_%s' % field]
                return
            for validation in onvalidation:
                validation(form)

        nextv = get_vars_next() or self.settings.request_reset_password_next
        if not self.settings.mailer:
            flash(self.messages.function_disabled)
            return ''
        onvalidation = self.settings.reset_password_onvalidation
        onaccept = self.settings.reset_password_onaccept
        log = self.messages['reset_password_log']
        rows = {}
        form = Form(
            self.form_data['request_reset_password'],
            hidden=dict(_next=nextv),
            submit=self.messages.password_reset_button,
            onvalidation=lambda form, rows=rows: process_form(form, rows)
        )
        if form.accepted:
            user = rows['user']
            if user.registration_key == 'pending':
                flash(self.auth.messages.registration_pending)
                redirect(self.auth.url('request_reset_password'))
            elif user.registration_key == 'blocked':
                flash(self.auth.messages.login_disabled)
                redirect(self.auth.url('login'))
            if self.auth.email_reset_password(user):
                flash(self.messages.email_sent)
            else:
                flash(self.messages.unable_to_send_email)
            self.auth.log_event(log, user)
            callback(onaccept, form)
            if not nextv:
                redirect(self.auth.url('request_reset_password'))
            else:
                nextv = replace_id(nextv, form)
            redirect(nextv)
        return form

    def retrieve_password(self):
        return self.request_reset_password()

    def change_password(self):
        def process_form(form):
            if form.params.old_password != row.password:
                form.errors.old_password = self.messages.invalid_password
                return
            if (
                form.params.new_password.password !=
                form.params.new_password2.password
            ):
                form.errors.new_password = self.messages.mismatched_password
                form.errors.new_password2 = self.messages.mismatched_password
                return
            for validation in onvalidation:
                validation(form)

        if not self.auth.is_logged_in():
            redirect(self.settings.login_url or self.auth.url('login'))
        row = self.auth.table_user[self.auth.user.id]
        nextv = get_vars_next() or self.settings.change_password_next
        onvalidation = self.settings.change_password_onvalidation
        onaccept = self.settings.change_password_onaccept
        log = self.messages['change_password_log']
        form = Form(
            self.form_data['change_password'],
            onvalidation=process_form,
            submit=self.messages.password_change_button,
            hidden=dict(_next=nextv)
        )
        if form.accepted:
            row.update(password=str(form.params.new_password))
            flash(self.messages.password_changed)
            self.auth.log_event(log, self.auth.user)
            callback(onaccept, form)
            if not nextv:
                nextv = self.auth.url('change_password')
            else:
                nextv = replace_id(nextv, form)
            redirect(nextv)
        return form

    def profile(self):
        if not self.auth.is_logged_in():
            redirect(self.settings.login_url)
        #passfield = self.settings.password_field
        nextv = get_vars_next() or self.settings.profile_next
        onvalidation = self.settings.profile_onvalidation
        onaccept = self.settings.profile_onaccept
        log = self.messages['profile_log']
        if not self.settings.profile_fields:
            profile_fields = [
                field.name for field in self.auth.table_user
                if field.type != 'password' and field.writable]
            self.settings.profile_fields = {
                'readable': profile_fields, 'writable': profile_fields}
        form = DALForm(
            self.auth.table_user,
            record_id=self.auth.user.id,
            fields=self.settings.profile_fields,
            hidden=dict(_next=nextv),
            submit=self.messages.profile_save_button,
            upload=self.settings.download_url,
            onvalidation=onvalidation
        )
        if form.accepted:
            self.auth.user.update(
                self.auth.table_user._filter_fields(form.params))
            flash(self.messages.profile_updated)
            self.auth.log_event(log, self.auth.user)
            callback(onaccept, form)
            # TODO: update this
            #if form.deleted:
            #   return self.logout()
            if not nextv:
                nextv = self.auth.url('profile')
            else:
                nextv = replace_id(nextv, form)
            redirect(nextv)
        return form

    """
    ## REMOVED in 0.4
    def groups(self):
        #: displays the groups and their roles for the logged in user
        if not self.is_logged_in():
            redirect(self.settings.login_url)
        memberships = self.db(
            self.table_membership.user_id == self.user.id).select()
        table = tag.table()
        for membership in memberships:
            groups = self.db(
                self.table_group.id == membership.group_id).select()
            if groups:
                group = groups[0]
                table.append(tag.tr(tag.h3(group.role, '(%s)' % group.id)))
                table.append(tag.tr(tag.p(group.description)))
        if not memberships:
            return None
        return table
    """

    def not_authorized(self):
        if request.isajax:
            abort(403, 'ACCESS DENIED')
        return 'ACCESS DENIED'



# -*- coding: utf-8 -*-
"""
    weppy._internal
    ---------------

    Provides internally used helpers and objects.

    :copyright: (c) 2014-2016 by Giovanni Barillari

    Several parts of this code comes from Flask and Werkzeug.
    :copyright: (c) 2014 by Armin Ronacher.

    :license: BSD, see LICENSE for more details.
"""

import os
import pkgutil
import sys
import warnings
from functools import partial
from ._compat import implements_iterator


def get_root_path(import_name):
    """Returns the path of the package or cwd if that cannot be found."""
    # Module already imported and has a file attribute.  Use that first.
    mod = sys.modules.get(import_name)
    if mod is not None and hasattr(mod, '__file__'):
        return os.path.dirname(os.path.abspath(mod.__file__))

    # Next attempt: check the loader.
    loader = pkgutil.get_loader(import_name)

    # Loader does not exist or we're referring to an unloaded main module
    # or a main module without path (interactive sessions), go with the
    # current working directory.
    if loader is None or import_name == '__main__':
        return os.getcwd()

    # For .egg, zipimporter does not have get_filename until Python 2.7.
    # Some other loaders might exhibit the same behavior.
    if hasattr(loader, 'get_filename'):
        filepath = loader.get_filename(import_name)
    else:
        # Fall back to imports.
        __import__(import_name)
        mod = sys.modules[import_name]
        filepath = getattr(mod, '__file__', None)

        # If we don't have a filepath it might be because we are a
        # namespace package.  In this case we pick the root path from the
        # first module that is contained in our package.
        if filepath is None:
            raise RuntimeError('No root path can be found for the provided '
                               'module "%s".  This can happen because the '
                               'module came from an import hook that does '
                               'not provide file name information or because '
                               'it\'s a namespace package.  In this case '
                               'the root path needs to be explicitly '
                               'provided.' % import_name)

    # filepath is import_name.py for a module, or __init__.py for a package.
    return os.path.dirname(os.path.abspath(filepath))


def create_missing_app_folders(app):
    try:
        for subfolder in ['languages', 'logs', 'static']:
            path = os.path.join(app.root_path, subfolder)
            if not os.path.exists(path):
                os.mkdir(path)
    except:
        pass


class ObjectProxy(object):
    #: Proxy to another object.
    __slots__ = ('__obj', '__dict__', '__name__')

    def __init__(self, obj, name=None):
        object.__setattr__(self, '_ObjectProxy__obj', obj)
        object.__setattr__(self, '__name__', name)

    def _get_robj(self):
        try:
            return getattr(self.__obj, self.__name__)
        except AttributeError:
            raise RuntimeError('no object bound to %s' % self.__name__)

    @property
    def __dict__(self):
        try:
            return self._get_robj().__dict__
        except RuntimeError:
            raise AttributeError('__dict__')

    def __repr__(self):
        try:
            obj = self._get_robj()
        except RuntimeError:
            return '<%s unbound>' % self.__class__.__name__
        return repr(obj)

    def __bool__(self):
        try:
            return bool(self._get_current_object())
        except RuntimeError:
            return False

    def __unicode__(self):
        try:
            return unicode(self._get_robj())
        except RuntimeError:
            return repr(self)

    def __dir__(self):
        try:
            return dir(self._get_current_object())
        except RuntimeError:
            return []

    def __getattr__(self, name):
        return getattr(self._get_robj(), name)

    def __setitem__(self, key, value):
        self._get_robj()[key] = value

    def __delitem__(self, key):
        del self._get_robj()[key]

    __setattr__ = lambda x, n, v: setattr(x._get_robj(), n, v)
    __delattr__ = lambda x, n: delattr(x._get_robj(), n)
    __str__ = lambda x: str(x._get_robj())
    __getitem__ = lambda x, i: x._get_robj()[i]
    __eq__ = lambda x, o: x._get_robj() == o
    __ne__ = lambda x, o: x._get_robj() != o
    __call__ = lambda x, *a, **kw: x._get_robj()(*a, **kw)
    __iter__ = lambda x: iter(x._get_robj())
    __contains__ = lambda x, i: i in x._get_robj()


@implements_iterator
class LimitedStream(object):
    def __init__(self, stream, limit):
        self._read = stream.read
        self._readline = stream.readline
        self._pos = 0
        self.limit = limit

    def __iter__(self):
        return self

    @property
    def is_exhausted(self):
        """If the stream is exhausted this attribute is `True`."""
        return self._pos >= self.limit

    def on_exhausted(self):
        """This is called when the stream tries to read past the limit.
        The return value of this function is returned from the reading
        function.
        """
        # Read null bytes from the stream so that we get the
        # correct end of stream marker.
        return self._read(0)

    def on_disconnect(self):
        raise Exception

    def exhaust(self, chunk_size=1024 * 64):
        """Exhaust the stream.  This consumes all the data left until the
        limit is reached.

        :param chunk_size: the size for a chunk.  It will read the chunk
                           until the stream is exhausted and throw away
                           the results.
        """
        to_read = self.limit - self._pos
        chunk = chunk_size
        while to_read > 0:
            chunk = min(to_read, chunk)
            self.read(chunk)
            to_read -= chunk

    def read(self, size=None):
        """Read `size` bytes or if size is not provided everything is read.

        :param size: the number of bytes read.
        """
        if self._pos >= self.limit:
            return self.on_exhausted()
        if size is None or size == -1:  # -1 is for consistence with file
            size = self.limit
        to_read = min(self.limit - self._pos, size)
        try:
            read = self._read(to_read)
        except (IOError, ValueError):
            return self.on_disconnect()
        if to_read and len(read) != to_read:
            return self.on_disconnect()
        self._pos += len(read)
        return read

    def readline(self, size=None):
        """Reads one line from the stream."""
        if self._pos >= self.limit:
            return self.on_exhausted()
        if size is None:
            size = self.limit - self._pos
        else:
            size = min(size, self.limit - self._pos)
        try:
            line = self._readline(size)
        except (ValueError, IOError):
            return self.on_disconnect()
        if size and not line:
            return self.on_disconnect()
        self._pos += len(line)
        return line

    def readlines(self, size=None):
        """Reads a file into a list of strings.  It calls :meth:`readline`
        until the file is read to the end.  It does support the optional
        `size` argument if the underlaying stream supports it for
        `readline`.
        """
        last_pos = self._pos
        result = []
        if size is not None:
            end = min(self.limit, last_pos + size)
        else:
            end = self.limit
        while 1:
            if size is not None:
                size -= last_pos - self._pos
            if self._pos >= end:
                break
            result.append(self.readline(size))
            if size is not None:
                last_pos = self._pos
        return result

    def __next__(self):
        line = self.readline()
        if not line:
            raise StopIteration()
        return line


@implements_iterator
class ClosingIterator(object):
    def __init__(self, iterable, callbacks=None):
        iterator = iter(iterable)
        self._next = partial(next, iterator)
        if callbacks is None:
            callbacks = []
        elif callable(callbacks):
            callbacks = [callbacks]
        else:
            callbacks = list(callbacks)
        iterable_close = getattr(iterator, 'close', None)
        if iterable_close:
            callbacks.insert(0, iterable_close)
        self._callbacks = callbacks

    def __iter__(self):
        return self

    def __next__(self):
        return self._next()

    def close(self):
        for callback in self._callbacks:
            callback()


class RemovedInNextVersionWarning(DeprecationWarning):
    pass


warnings.simplefilter('always', RemovedInNextVersionWarning)


def warn_of_deprecation(old_name, new_name, prefix=None, stack=2):
    msg = "%(old)s is deprecated, use %(new)s instead."
    if prefix:
        msg = "%(prefix)s." + msg
    warnings.warn(
        msg % {'old': old_name, 'new': new_name, 'prefix': prefix},
        RemovedInNextVersionWarning, stack)


class deprecated(object):
    def __init__(self, old_method_name, new_method_name, class_name=None, s=0):
        self.class_name = class_name
        self.old_method_name = old_method_name
        self.new_method_name = new_method_name
        self.additional_stack = s

    def __call__(self, f):
        def wrapped(*args, **kwargs):
            warn_of_deprecation(
                self.old_method_name, self.new_method_name, self.class_name,
                3 + self.additional_stack)
            return f(*args, **kwargs)
        return wrapped



# -*- coding: utf-8 -*-
"""
    weppy.expose
    ------------

    Provide routing for application, as well as internal url creation.
    The expose module is responsible of dispatching the app dynamic requests.

    :copyright: (c) 2014-2016 by Giovanni Barillari
    :license: BSD, see LICENSE for more details.
"""

import re
import os
from collections import OrderedDict
from ._compat import PY2, with_metaclass, itervalues, iteritems, text_type
from .handlers import Handler, _wrapWithHandlers
from .templating.helpers import TemplateMissingError
from .globals import current
from .http import HTTP

if PY2:
    from urllib import quote as uquote
else:
    from urllib.parse import quote as uquote

__all__ = ['Expose', 'url']


class MetaExpose(type):
    def __new__(cls, name, bases, attrs):
        nc = type.__new__(cls, name, bases, attrs)
        nc._get_routes_in_for_host = nc._get_routes_in_for_host_simple
        return nc


class Expose(with_metaclass(MetaExpose)):
    _routing_stack = []
    _routes_str = OrderedDict()
    application = None
    routes_in = {'__any__': OrderedDict()}
    routes_out = {}
    common_handlers = []
    common_helpers = []
    processors = []
    REGEX_INT = re.compile('<int\:(\w+)>')
    REGEX_STR = re.compile('<str\:(\w+)>')
    REGEX_ANY = re.compile('<any\:(\w+)>')
    REGEX_ALPHA = re.compile('<alpha\:(\w+)>')
    REGEX_DATE = re.compile('<date\:(\w+)>')
    REGEX_DECORATION = re.compile(
        '(([?*+])|(\([^()]*\))|(\[[^\[\]]*\])|(\<[^<>]*\>))')

    def __init__(self, path=None, name=None, template=None, handlers=None,
                 helpers=None, schemes=None, hostname=None, methods=None,
                 prefix=None, template_folder=None, template_path=None):
        if callable(path):
            raise SyntaxError('Use @route(), not @route.')
        self.schemes = schemes or ('http', 'https')
        if not isinstance(self.schemes, (list, tuple)):
            self.schemes = (self.schemes,)
        self.methods = methods or ('get', 'post', 'head')
        if not isinstance(self.methods, (list, tuple)):
            self.methods = (self.methods,)
        self.hostname = hostname
        self.path = path
        self.name = name
        self.template = template
        self.template_folder = template_folder
        self.template_path = template_path
        self.prefix = prefix
        self.handlers = [_ResponseHandler(self)] + self.common_handlers + \
            (handlers or []) + self.common_helpers + (helpers or [])
        # check handlers are indeed valid handlers
        if any(not isinstance(handler, Handler) for handler in self.handlers):
            raise RuntimeError('Invalid Handler')
        self._routing_stack.append(self)

    @property
    def folder(self):
        return self.application.root_path

    def build_name(self):
        short = self.filename[1 + len(self.folder):].rsplit('.', 1)[0]
        if not short:
            short = self.filename.rsplit('.', 1)[0]
        if short == "__init__":
            short = self.folder.rsplit('/', 1)[-1]
        # allow only one level of naming if name is builded
        if len(short.split(os.sep)) > 1:
            short = short.split(os.sep)[-1]
        return '.'.join(short.split(os.sep) + [self.func_name])

    @classmethod
    def build_regex(cls, schemes, hostname, methods, path):
        path = cls.REGEX_INT.sub('(?P<\g<1>>\d+)', path)
        path = cls.REGEX_STR.sub('(?P<\g<1>>[^/]+)', path)
        path = cls.REGEX_ANY.sub('(?P<\g<1>>.*)', path)
        path = cls.REGEX_ALPHA.sub('(?P<\g<1>>[^/\W\d_]+)', path)
        path = cls.REGEX_DATE.sub('(?P<\g<1>>\d{4}-\d{2}-\d{2})', path)
        re_schemes = ('|'.join(schemes)).lower()
        re_methods = ('|'.join(methods)).lower()
        re_hostname = re.escape(hostname) if hostname else '[^/]*'
        expr = '^(%s) (%s)\://(%s)(%s)$' % \
            (re_methods, re_schemes, re_hostname, path)
        return expr

    @staticmethod
    def remove_trailslash(path):
        if path.endswith("/") and len(path) > 1:
            return path[:-1]
        return path

    @staticmethod
    def override_midargs(path):
        args = path.split("//")
        if len(args) > 1:
            path = "/{{:arg:}}/".join(args)
        return path

    @classmethod
    def remove_decoration(cls, path):
        """
        converts somehing like "/junk/test_args/<str:a>(/<int:b>)?"
        into something like    "/junk/test_args" for reverse routing
        """
        while True:
            new_path = cls.REGEX_DECORATION.sub('', path)
            new_path = cls.remove_trailslash(new_path)
            new_path = cls.override_midargs(new_path)
            if new_path == path:
                return path
            else:
                path = new_path

    @classmethod
    def add_route(cls, route):
        host = route[1].hostname or '__any__'
        if host not in cls.routes_in:
            cls.routes_in[host] = OrderedDict()
            cls._get_routes_in_for_host = cls._get_routes_in_for_host_all
        cls.routes_in[host][route[1].name] = route
        cls.routes_out[route[1].name] = {
            'host': route[1].hostname,
            'path': cls.remove_decoration(route[1].path)}

    def __call__(self, func):
        self.func_name = func.__name__
        self.filename = os.path.realpath(func.__code__.co_filename)
        self.hostname = self.hostname or \
            self.application.config.hostname_default
        if not self.path:
            self.path = '/' + func.__name__
        if not self.name:
            self.name = self.build_name()
        # is it good?
        if self.name.endswith("."):
            self.name = self.name + self.func_name
        #
        if not self.path.startswith('/'):
            self.path = '/' + self.path
        if self.prefix:
            if not self.prefix.startswith('/'):
                self.prefix = '/' + self.prefix
            self.path = (self.path != '/' and self.prefix + self.path) \
                or self.prefix
        if not self.template:
            self.template = self.func_name + \
                self.application.template_default_extension
        if self.template_folder:
            self.template = os.path.join(self.template_folder, self.template)
        self.template_path = self.template_path or \
            self.application.template_path
        wrapped_func = _wrapWithHandlers(self.handlers)(func)
        self.func = wrapped_func
        self.regex = self.build_regex(
            self.schemes, self.hostname, self.methods, self.path)
        route = (re.compile(self.regex), self)
        self.add_route(route)
        self._routes_str[self.name] = "%s %s://%s%s -> %s" % (
            "|".join(s.upper() for s in self.methods),
            "|".join(s for s in self.schemes),
            self.hostname or "<any>",
            self.path,
            self.name
        )
        for proc_handler in self.processors:
            proc_handler(self)
        self._routing_stack.pop()
        return func

    @classmethod
    def match_lang(cls, path):
        default = cls.application.language_default
        if len(path) <= 1:
            return path, default
        clean_path = path.lstrip('/')
        lang = clean_path.split('/', 1)[0]
        if lang in cls.application.languages and lang != default:
            new_path = '/'.join([arg for arg in clean_path.split('/')[1:]])
            if path.startswith('/'):
                new_path = '/' + new_path
            return new_path, lang
        return path, default

    @classmethod
    def _get_routes_in_for_host_all(cls, hostname):
        return (
            cls.routes_in.get(hostname, cls.routes_in['__any__']),
            cls.routes_in['__any__'])

    @classmethod
    def _get_routes_in_for_host_simple(cls, hostname):
        return (cls.routes_in['__any__'],)

    @classmethod
    def match(cls, request):
        path = cls.remove_trailslash(request.path_info)
        if cls.application.language_force_on_url:
            path, lang = cls.match_lang(path)
            request.language = lang
            current._language = request.language
        else:
            request.language = None
        expression = '%s %s://%s%s' % (
            request.method, request.scheme, request.hostname, path)
        for routes in cls._get_routes_in_for_host(request.hostname):
            for regex, obj in itervalues(routes):
                match = regex.match(expression)
                if match:
                    return obj, match.groupdict()
        return None, {}

    @staticmethod
    def _after_dispatch(route):
        #: call handlers `on_end` method
        for handler in reversed(route.handlers):
            handler.on_end()

    @classmethod
    def dispatch(cls):
        #: get the right exposed function
        request = current.request
        route, reqargs = cls.match(request)
        if route:
            request.name = route.name
            try:
                route.func(**reqargs)
            except:
                cls._after_dispatch(route)
                raise
        else:
            raise HTTP(404, body="Invalid action\n")
        #: end the dispatching
        cls._after_dispatch(route)

    @classmethod
    def static_versioning(cls):
        return (cls.application.config.static_version_urls and
                cls.application.config.static_version) or ''

    @classmethod
    def exposing(cls):
        return cls._routing_stack[-1]


class _ResponseHandler(Handler):
    def __init__(self, route):
        self.route = route

    def wrap_call(self, func):
        def wrap(*args, **kwargs):
            response = current.response
            output = func(*args, **kwargs)
            if output is None:
                output = {'current': current, 'url': url}
            if isinstance(output, dict):
                output['current'] = output.get('current', current)
                output['url'] = output.get('url', url)
                try:
                    output = Expose.application.templater.render(
                        self.route.template_path, self.route.template, output)
                except TemplateMissingError:
                    raise HTTP(404, body="Invalid view\n")
                response.output = output
            elif isinstance(output, text_type) or hasattr(output, '__iter__'):
                response.output = output
            else:
                response.output = str(output)
        return wrap


def url(path, args=[], params={}, extension=None, sign=None, scheme=None,
        host=None, language=None):
    """
    usages:
        url('index') # assumes app or default expose file
        url('.index') # use current exposed file
        url('mod.index') # index function in 'mod' module
        url('static', 'file') # for static files
        url('/myurl') # a normal url
    """

    if not isinstance(args, (list, tuple)):
        args = [args]
    # allow user to use url('static', 'file')
    if path == 'static':
        path = '/static'
    # routes urls with 'dot' notation
    if '/' not in path:
        # urls like 'function' refers to same module
        if '.' not in path:
            namespace = Expose.application.config.url_default_namespace or \
                Expose.application.name
            path = namespace + "." + path
        # urls like '.function' refers to main app module
        elif path.startswith('.'):
            if not hasattr(current, 'request'):
                raise RuntimeError(
                    'cannot build url("%s",...) without current request' % path
                )
            module = current.request.name.rsplit('.', 1)[0]
            path = module + path
        # find correct route
        try:
            url = Expose.routes_out[path]['path']
            url_host = Expose.routes_out[path]['host']
            # try to rebuild url if midargs found
            midargs = url.split("{{:arg:}}")
            if len(midargs) > 1:
                u = ""
                if len(args) >= len(midargs) - 1:
                    for i in range(0, len(midargs) - 1):
                        u += midargs[i] + uquote(str(args[i]))
                    u += midargs[-1]
                    url = u
                    args = args[len(midargs) - 1:]
                else:
                    raise RuntimeError(
                        'invalid url("%s",...): needs args for params' % path
                    )
            # try to use the correct hostname
            if url_host is not None:
                try:
                    if current.request.hostname != url_host:
                        #url = current.request.scheme+"://"+url_host+url
                        scheme = current.request.scheme
                        host = url_host
                except:
                    pass
        except KeyError:
            raise RuntimeError('invalid url("%s",...)' % path)
    # handle classic urls
    else:
        url = path
    # add static versioning
    if url[0:7] == '/static':
        if Expose.static_versioning():
            url = url[0:7] + "/_" + str(Expose.static_versioning()) + url[7:]
    # language
    if Expose.application.language_force_on_url:
        if url.startswith("/"):
            lang = None
            if language:
                #: use the given language if is enabled in application
                if language in Expose.application.languages:
                    lang = language
            else:
                #: try to use the request language if context exists
                if hasattr(current, 'request'):
                    lang = current.request.language
            if lang and lang != Expose.application.language_default:
                url = '/' + lang + url
    # add extension (useless??)
    if extension:
        url = url + '.' + extension
    # add args
    if args:
        if not isinstance(args, (list, tuple)):
            args = (args,)
        url = url + '/' + '/'.join(uquote(str(a)) for a in args)
    # add signature
    if sign:
        params['_signature'] = sign(url)
    # add params
    if params:
        url = url + '?' + '&'.join(
            '%s=%s' % (
                uquote(str(k)), uquote(str(v))) for k, v in iteritems(params)
        )
    # scheme=True means to use current scheme
    if scheme is True:
        if not hasattr(current, 'request'):
            raise RuntimeError(
                'cannot build url("%s",...) without current request' % path
            )
        scheme = current.request.scheme
    # add scheme and host
    if scheme:
        if host is None:
            if not hasattr(current, 'request'):
                raise RuntimeError(
                    'cannot build url("%s",...) without current request' % path
                )
            host = current.request.hostname
        url = '%s://%s%s' % (scheme, host, url)
    return url



# -*- coding: utf-8 -*-
"""
    weppy.cli
    ---------

    Provide command line tools for weppy applications.

    :copyright: (c) 2014-2016 by Giovanni Barillari

    Based on the code of Flask (http://flask.pocoo.org)
    :copyright: (c) 2014 by Armin Ronacher.

    :license: BSD, see LICENSE for more details.
"""

from __future__ import print_function

import click
import os
import sys
import types
from ._compat import itervalues, iteritems
from . import __version__ as weppy_version


def find_best_app(module):
    """Given a module instance this tries to find the best possible
    application in the module or raises an exception.
    """
    from .app import App

    # Search for the most common names first.
    for attr_name in 'app', 'application':
        app = getattr(module, attr_name, None)
        if app is not None and isinstance(app, App):
            return app

    # Otherwise find the only object that is a weppy App instance.
    matches = [v for k, v in iteritems(module.__dict__) if isinstance(v, App)]

    if len(matches) == 1:
        return matches[0]
    raise Exception(
        'Failed to find application in module "%s".' % module.__name__
    )


def find_dal(module, var_name=None):
    #: Given a module instance this tries to find the dal instances
    #  in the module.
    if var_name:
        return [getattr(module, var_name)]
    from .dal import DAL
    matches = [v for k, v in iteritems(module.__dict__) if isinstance(v, DAL)]

    return matches


def prepare_exec_for_file(filename):
    """Given a filename this will try to calculate the python path, add it
    to the search path and return the actual module name that is expected.
    """
    module = []

    # Chop off file extensions or package markers
    if filename.endswith('.py'):
        filename = filename[:-3]
    elif os.path.split(filename)[1] == '__init__.py':
        filename = os.path.dirname(filename)
    else:
        raise Exception(
            'The file provided (%s) does is not a valid Python file.')
    filename = os.path.realpath(filename)

    dirpath = filename
    while 1:
        dirpath, extra = os.path.split(dirpath)
        module.append(extra)
        if not os.path.isfile(os.path.join(dirpath, '__init__.py')):
            break

    sys.path.insert(0, dirpath)
    return '.'.join(module[::-1])


def get_app_module(app_id):
    if ':' in app_id:
        module, app_obj = app_id.split(':', 1)
    else:
        module = app_id
        app_obj = None
    mod = sys.modules.get(module)
    if mod is None:
        __import__(module)
        mod = sys.modules[module]
    return module, mod, app_obj


def locate_app(app_id):
    """Attempts to locate the application."""
    module, mod, app_obj = get_app_module(app_id)
    if app_obj is None:
        app = find_best_app(mod)
    else:
        app = getattr(mod, app_obj, None)
        if app is None:
            raise RuntimeError('Failed to find application in module "%s"'
                               % module)
    return app


class ScriptInfo(object):
    """Help object to deal with Weppy applications.  This is usually not
    necessary to interface with as it's used internally in the dispatching
    to click.
    """

    def __init__(self, app_import_path=None, debug=None):
        #: The application import path
        self.app_import_path = app_import_path
        #: The debug flag.  If this is not None, the application will
        #: automatically have it's debug flag overridden with this value.
        self.debug = debug
        #: A dictionary with arbitrary data that can be associated with
        #: this script info.
        self.data = {}
        self._loaded_ctx = None
        self._loaded_app = None
        self.db_var_name = None

    def load_appctx(self):
        if self._loaded_ctx is not None:
            return self._loaded_ctx
        if self.app_import_path is None:
            raise Exception("Could not locate application.")
        module, mod, app_obj = get_app_module(self.app_import_path)
        ctx = {}
        for key, value in iteritems(mod.__dict__):
            if key == "__builtins__" or isinstance(value, types.FunctionType):
                continue
            ctx[key] = value
        self._loaded_ctx = ctx
        return ctx

    def load_app(self):
        """Loads the app (if not yet loaded) and returns it.  Calling
        this multiple times will just result in the already loaded app to
        be returned.
        """
        if self._loaded_app is not None:
            return self._loaded_app
        if self.app_import_path is None:
            raise Exception("Could not locate application.")
        a = locate_app(self.app_import_path)
        if self.debug is not None:
            a.debug = self.debug
        self._loaded_app = a
        return a

    def load_dal(self):
        if self.app_import_path is None:
            raise Exception("Could not locate application.")
        module, mod, app_obj = get_app_module(self.app_import_path)
        return find_dal(mod, self.db_var_name)


pass_script_info = click.make_pass_decorator(ScriptInfo)


def set_app_value(ctx, param, value):
    if value is not None:
        if os.path.isfile(value):
            value = prepare_exec_for_file(value)
        elif '.' not in sys.path:
            sys.path.insert(0, '.')
    ctx.ensure_object(ScriptInfo).app_import_path = value


app_option = click.Option(
    ['-a', '--app'],
    help='The application to run',
    callback=set_app_value, is_eager=True
)


class WeppyGroup(click.Group):
    def __init__(self, add_default_commands=True, add_app_option=True,
                 add_debug_option=True, **extra):
        params = list(extra.pop('params', None) or ())
        if add_app_option:
            params.append(app_option)
        #if add_debug_option:
        #    params.append(debug_option)

        click.Group.__init__(self, params=params, **extra)
        #self.create_app = create_app

        if add_default_commands:
            self.add_command(run_command)
            self.add_command(shell_command)
            self.add_command(routes_command)

    def get_command(self, ctx, name):
        # We load built-in commands first as these should always be the
        # same no matter what the app does.  If the app does want to
        # override this it needs to make a custom instance of this group
        # and not attach the default commands.
        #
        # This also means that the script stays functional in case the
        # application completely fails.
        rv = click.Group.get_command(self, ctx, name)
        if rv is not None:
            return rv

        info = ctx.ensure_object(ScriptInfo)
        try:
            rv = info.load_app().cli.get_command(ctx, name)
            if rv is not None:
                return rv
        except:
            pass

    def main(self, *args, **kwargs):
        obj = kwargs.get('obj')
        if obj is None:
            obj = ScriptInfo()
        kwargs['obj'] = obj
        return click.Group.main(self, *args, **kwargs)


@click.command('run', short_help='Runs a development server.')
@click.option('--host', '-h', default='127.0.0.1',
              help='The interface to bind to.')
@click.option('--port', '-p', default=8000,
              help='The port to bind to.')
@click.option('--reloader', type=(bool), default=True,
              help='Runs with reloader.')
@click.option('--debug', type=(bool), default=True, help='Runs in debug mode.')
@pass_script_info
def run_command(info, host, port, reloader, debug):
    app = info.load_app()
    app.debug = debug
    if os.environ.get('WEPPY_RUN_MAIN') != 'true':
        print("> Serving weppy application %s" % app.import_name)
        quit_msg = "(press CTRL+C to quit)"
        print("> weppy application %s running on http://%s:%i %s" %
              (app.import_name, host, port, quit_msg))
    if reloader:
        from ._reloader import run_with_reloader
        run_with_reloader(app, host, port)
    else:
        app._run(host, port)


@click.command('shell', short_help='Runs a shell in the app context.')
@pass_script_info
def shell_command(info):
    import code
    ctx = info.load_appctx()
    app = info.load_app()
    banner = 'Python %s on %s\nweppy %s shell on app: %s' % (
        sys.version,
        sys.platform,
        weppy_version,
        app.import_name
    )
    code.interact(banner=banner, local=app.make_shell_context(ctx))


@click.command('routes', short_help='Display the app routing table.')
@pass_script_info
def routes_command(info):
    app = info.load_app()
    print("> Routing table for weppy application %s:" % app.import_name)
    for route in itervalues(app.route._routes_str):
        print(route)


cli = WeppyGroup(help="")


def set_db_value(ctx, param, value):
    ctx.ensure_object(ScriptInfo).db_var_name = value


db_option = click.Option(
    ['--db'],
    help='The db instance to use',
    callback=set_db_value, is_eager=True
)


@cli.group('migrations', short_help='Runs migration operations.')
def migrations_cli():
    pass


@migrations_cli.command(
    'status', short_help='Shows current database revision.')
@click.option('--verbose', '-v', default=False, is_flag=True)
@pass_script_info
def migrations_status(info, verbose):
    from .dal.migrations.commands import status
    app = info.load_app()
    dals = info.load_dal()
    status(app, dals, verbose)


@migrations_cli.command('history', short_help="Shows migrations history.")
@click.option('--range', '-r', default=None)
@click.option('--verbose', '-v', default=False, is_flag=True)
@pass_script_info
def migrations_history(info, range, verbose):
    from .dal.migrations.commands import history
    app = info.load_app()
    dals = info.load_dal()
    history(app, dals, range, verbose)


@migrations_cli.command(
    'generate', short_help='Generate a new migration from application models.')
@click.option('--message', '-m', default='Generated migration',
              help='The description for the new migration.')
@click.option('-head', default='head', help='The migration to generate from')
@pass_script_info
def migrations_generate(info, message, head):
    from .dal.migrations.commands import generate
    app = info.load_app()
    dals = info.load_dal()
    generate(app, dals, message, head)


@migrations_cli.command('new', short_help='Generate a new empty migration.')
@click.option('--message', '-m', default='New migration',
              help='The description for the new migration.')
@click.option('-head', default='head', help='The migration to generate from')
@pass_script_info
def migrations_new(info, message, head):
    from .dal.migrations.commands import new
    app = info.load_app()
    dals = info.load_dal()
    new(app, dals, message, head)


@migrations_cli.command(
    'up', short_help='Upgrades the database to the selected migration.')
@click.option('--revision', '-r', default='head',
              help='The migration to upgrade to.')
@pass_script_info
def migrations_up(info, revision):
    from .dal.migrations.commands import up
    app = info.load_app()
    dals = info.load_dal()
    up(app, dals, revision)


@migrations_cli.command(
    'down', short_help='Downgrades the database to the selected migration.')
@click.option('--revision', '-r', required=True,
              help='The migration to downgrade to.')
@pass_script_info
def migrations_down(info, revision):
    from .dal.migrations.commands import down
    app = info.load_app()
    dals = info.load_dal()
    down(app, dals, revision)


def main(as_module=False):
    this_module = __package__ + '.cli'
    args = sys.argv[1:]

    if as_module:
        if sys.version_info >= (2, 7):
            name = 'python -m ' + this_module.rsplit('.', 1)[0]
        else:
            name = 'python -m ' + this_module

        # This module is always executed as "python -m run" and as such
        # we need to ensure that we restore the actual command line so that
        # the reloader can properly operate.
        sys.argv = ['-m', this_module] + sys.argv[1:]
    else:
        name = None

    cli.main(args=args, prog_name=name)

if __name__ == '__main__':
    main(as_module=True)



__version__ = '0.8.5'

from .app import App, AppModule
from .expose import url
from .http import redirect
from .helpers import abort, stream_file
from .tags import tag, cat, safe, asis
from .forms import Form, DALForm
from .cache import Cache
from .dal import DAL, Field
from .globals import request, response, session
from .handlers import Handler, Helper
from .language import T
from .datastructures import sdict



# -*- coding: utf-8 -*-
"""
    tests.auth
    ----------

    Test weppy Auth module

    :copyright: (c) 2014-2016 by Giovanni Barillari
    :license: BSD, see LICENSE for more details.
"""

import pytest
from weppy import App, sdict
from weppy.dal import DAL, Field, Model, has_many, belongs_to
from weppy.tools import Auth
from weppy.tools.auth.models import AuthUser


class User(AuthUser):
    has_many('things')
    gender = Field()


class Thing(Model):
    belongs_to('user')


@pytest.fixture(scope='module')
def db():
    app = App(__name__)
    db = DAL(app, config=sdict(uri='sqlite:memory'))
    auth = Auth(app, db, usermodel=User)
    db.define_models(Thing)
    return db


def test_models(db):
    user = User.create(
        email="walter@massivedynamics.com",
        password="pocketuniverse",
        first_name="Walter",
        last_name="Bishop"
    )
    group = db.auth_groups.insert(
        role="admin"
    )
    group2 = db.auth_groups.insert(
        role="moderator"
    )
    db.auth_memberships.insert(
        user=user.id,
        authgroup=group
    )
    db.auth_permissions.insert(
        authgroup=group
    )
    user = db.users[1]
    assert len(user.authmemberships()) == 1
    assert user.authmemberships()[0].user == 1
    assert len(user.authgroups()) == 1
    assert user.authgroups()[0].role == "admin"
    assert len(user.authpermissions()) == 1
    assert user.authpermissions()[0].authgroup.role == "admin"
    assert len(group.authmemberships()) == 1
    assert group.authmemberships()[0].authgroup == 1
    assert len(group.users()) == 1
    assert group.users()[0].email == "walter@massivedynamics.com"
    assert len(group.authpermissions()) == 1
    user.authgroups.add(group2)
    assert len(user.authgroups(db.auth_groups.id)) == 2
    assert user.authgroups(db.auth_groups.role)[1].role == "moderator"
    assert len(user.things()) == 0



# -*- coding: utf-8 -*-
"""
    weppy.language.helpers
    ----------------------

    Adapted from the web2py's code (http://www.web2py.com)

    :copyright: (c) by Massimo Di Pierro <mdipierro@cs.depaul.edu>
    :license: LGPLv3 (http://www.gnu.org/licenses/lgpl.html)
"""

import ast
import os
import pkgutil
import re

from .._compat import PY2, to_unicode, to_bytes
from .cache import clear_cache, getcfs
from ..libs.portalocker import read_locked, LockedFile


regex_backslash = re.compile(r"\\([\\{}%])")
regex_langfile = re.compile('^[a-z]{2}(-[a-z]{2})?\.py$')
regex_plural = re.compile('%({.+?})')
# %%{word(varname or number)}
regex_plural_dict = re.compile(
    '^{(?P<w>[^()[\]][^()[\]]*?)\((?P<n>[^()\[\]]+)\)}$')
# %%{word[index]} or %%{word}
regex_plural_tuple = re.compile('^{(?P<w>[^[\]()]+)(?:\[(?P<i>\d+)\])?}$')
regex_plural_file = re.compile('^plural-[a-zA-Z]{2}(-[a-zA-Z]{2})?\.py$')
regex_language = re.compile(
    '([a-z]{2,3}(?:\-[a-z]{2})?(?:\-[a-z]{2})?)(?:[,;]|$)')


DEFAULT_LANGUAGE = 'en'
DEFAULT_LANGUAGE_NAME = 'English'

# DEFAULT PLURAL-FORMS RULES:
# language doesn't use plural forms
DEFAULT_NPLURALS = 1
# only one singular/plural form is used
DEFAULT_GET_PLURAL_ID = lambda n: 0
# word is unchangeable
DEFAULT_CONSTRUCT_PLURAL_FORM = lambda word, plural_id: word


def read_possible_plural_rules():
    """
    create list of all possible plural rules files
    result is cached in PLURAL_RULES dictionary to increase speed
    """
    plurals = {}
    try:
        import weppy.language.plurals as package
        for importer, modname, ispkg in pkgutil.iter_modules(package.__path__):
            if len(modname) == 2:
                module = __import__(package.__name__ + '.' + modname,
                                    fromlist=[modname])
                lang = modname
                #pname = modname + '.py'
                nplurals = getattr(module, 'nplurals', DEFAULT_NPLURALS)
                get_plural_id = getattr(
                    module, 'get_plural_id',
                    DEFAULT_GET_PLURAL_ID)
                construct_plural_form = getattr(
                    module, 'construct_plural_form',
                    DEFAULT_CONSTRUCT_PLURAL_FORM)
                plurals[lang] = (lang, nplurals, get_plural_id,
                                 construct_plural_form)
    except ImportError:
        pass
    return plurals

PLURAL_RULES = read_possible_plural_rules()


def read_plural_dict_aux(filename):
    lang_text = read_locked(filename).replace('\r\n', '\n')
    try:
        return eval(lang_text) or {}
    except Exception:
        #e = sys.exc_info()[1]
        #status = 'Syntax error in %s (%s)' % (filename, e)
        #logging.error(status)
        status = 'Syntax error in %s' % filename
        return {'__corrupted__': status}


def read_plural_dict(filename):
    return getcfs('plurals:' + filename, filename,
                  lambda: read_plural_dict_aux(filename))


def write_plural_dict(filename, contents):
    if '__corrupted__' in contents:
        return
    try:
        fp = LockedFile(filename, 'wb')
        fp.write(to_bytes(u'#!/usr/bin/env python\n{\n# "singular form (0)": ["first plural form (1)", "second plural form (2)", ...],\n'))
        # coding: utf8\n{\n')
        for key in sorted(contents):
            forms = u'[' + u','.join('"' + form + '"' for form in contents[key]) + u']'
            val = u'"%s": %s,\n' % (key, forms)
            fp.write(to_bytes(val))
        fp.write(to_bytes(u'}\n'))
    except (IOError, OSError):
        #if not is_gae:
        #    logging.warning('Unable to write to file %s' % filename)
        return
    finally:
        fp.close()


def safe_eval(text):
    if text.strip():
        try:
            return ast.literal_eval(text)
        except ImportError:
            return eval(text, {}, {})
    return None


def read_dict_aux(filename):
    lang_text = read_locked(filename).replace('\r\n', '\n')
    clear_cache(filename)
    try:
        rv = safe_eval(lang_text) or {}
        if PY2:
            for key, val in rv.items():
                rv[to_unicode(key)] = to_unicode(val)
    except Exception:
        #e = sys.exc_info()[1]
        #status = 'Syntax error in %s (%s)' % (filename, e)
        #logging.error(status)
        status = 'Syntax error in %s' % filename
        rv = {'__corrupted__': status}
    return rv


def read_dict(filename):
    """ return dictionary with translation messages
    """
    return getcfs('lang:' + filename, filename,
                  lambda: read_dict_aux(filename))


def write_dict(filename, contents):
    if '__corrupted__' in contents:
        return
    try:
        fp = LockedFile(filename, 'wb')
    except (IOError, OSError):
        return
    fp.write(to_bytes(u'# coding: utf8\n{\n'))
    for key in sorted(contents):
        val = u'"%s": "%s",\n' % (key, contents[key])
        fp.write(to_bytes(val))
    fp.write(to_bytes(u'}\n'))
    fp.close()


def read_possible_languages_aux(langdir):
    def get_lang_struct(lang, langcode, langname, langfile_mtime):
        if lang == 'default':
            real_lang = langcode.lower()
        else:
            real_lang = lang
        (prules_langcode,
         nplurals,
         get_plural_id,
         construct_plural_form
         ) = PLURAL_RULES.get(real_lang[:2], ('default',
                                              DEFAULT_NPLURALS,
                                              DEFAULT_GET_PLURAL_ID,
                                              DEFAULT_CONSTRUCT_PLURAL_FORM))
        if prules_langcode != 'default':
            (pluraldict_fname, pluraldict_mtime) = plurals.get(
                real_lang,
                plurals.get(real_lang[:2], ('plural-%s.py' % real_lang, 0)))
        else:
            pluraldict_fname = None
            pluraldict_mtime = 0
        return (langcode,        # language code from !langcode!
                langname,
                # language name in national spelling from !langname!
                langfile_mtime,  # m_time of language file
                pluraldict_fname,  # name of plural dictionary file or None (when default.py is not exist)
                pluraldict_mtime,  # m_time of plural dictionary file or 0 if file is not exist
                prules_langcode,  # code of plural rules language or 'default'
                nplurals,        # nplurals for current language
                get_plural_id,   # get_plural_id() for current language
                construct_plural_form)  # construct_plural_form() for current language

    plurals = {}
    flist = os.listdir(langdir)
    # scan languages directory for plural dict files:
    for pname in flist:
        if regex_plural_file.match(pname):
            plurals[pname[7:-3]] = (
                pname,
                os.stat(os.path.join(langdir, pname)).st_mtime)
    langs = {}
    # scan languages directory for langfiles:
    for fname in flist:
        if regex_langfile.match(fname) or fname == 'default.py':
            fname_with_path = os.path.join(langdir, fname)
            d = read_dict(fname_with_path)
            lang = fname[:-3]
            langcode = d.get('!langcode!', lang if lang != 'default'
                             else DEFAULT_LANGUAGE)
            langname = d.get('!langname!', langcode)
            langfile_mtime = os.stat(fname_with_path).st_mtime
            langs[lang] = get_lang_struct(lang, langcode,
                                          langname, langfile_mtime)
    if 'default' not in langs:
        # if default.py is not found,
        # add DEFAULT_LANGUAGE as default language:
        langs['default'] = get_lang_struct('default', DEFAULT_LANGUAGE,
                                           DEFAULT_LANGUAGE_NAME, 0)
    deflang = langs['default']
    deflangcode = deflang[0]
    if deflangcode not in langs:
        # create language from default.py:
        langs[deflangcode] = deflang[:2] + (0,) + deflang[3:]

    return langs


def read_possible_languages(langpath):
    return getcfs('langs:' + langpath, langpath,
                  lambda: read_possible_languages_aux(langpath))


def upper_fun(s):
    return to_unicode(s).upper().encode('utf-8')


def title_fun(s):
    return to_unicode(s).title().encode('utf-8')


def cap_fun(s):
    return to_unicode(s).capitalize().encode('utf-8')


def _make_ttabin():
    ltrans = u"\\%{}"
    rtrans = u'\x1c\x1d\x1e\x1f'
    return dict((ord(char), rtrans) for char in ltrans)


def _make_ttabout():
    ltrans = u'\x1c\x1d\x1e\x1f'
    rtrans = u"\\%{}"
    return dict((ord(char), rtrans) for char in ltrans)


#ttab_in = maketrans(u"\\%{}", u'\x1c\x1d\x1e\x1f')
#ttab_out = maketrans(u'\x1c\x1d\x1e\x1f', u"\\%{}")
ttab_in = _make_ttabin()
ttab_out = _make_ttabout()



# -*- coding: utf-8 -*-
"""
    weppy.utils
    -----------

    Provides some utilities for weppy.

    :copyright: (c) 2014-2016 by Giovanni Barillari
    :license: BSD, see LICENSE for more details.
"""

import re
import socket


class cachedprop(object):
    #: a read-only @property that is only evaluated once.
    def __init__(self, fget, doc=None):
        self.fget = fget
        self.__doc__ = doc or fget.__doc__
        self.__name__ = fget.__name__

    def __get__(self, obj, cls):
        if obj is None:
            return self
        obj.__dict__[self.__name__] = result = self.fget(obj)
        return result


def is_valid_ip_address(address):
    REGEX_IPv4 = re.compile('(\d+)\.(\d+)\.(\d+)\.(\d+)')
    # deal with special cases
    if address.lower() in ('127.0.0.1', 'localhost', '::1',
                           '::ffff:127.0.0.1'):
        return True
    elif address.lower() in ('unknown', ''):
        return False
    elif address.count('.') == 3:  # assume IPv4
        if address.startswith('::ffff:'):
            address = address[7:]
        if hasattr(socket, 'inet_aton'):  # try validate using the OS
            try:
                socket.inet_aton(address)
                return True
            except socket.error:  # invalid address
                return False
        else:  # try validate using Regex
            match = REGEX_IPv4.match(address)
            if match and all(0 <= int(match.group(i)) < 256
                             for i in (1, 2, 3, 4)):
                return True
            return False
    elif hasattr(socket, 'inet_pton'):  # assume IPv6, try using the OS
        try:
            socket.inet_pton(socket.AF_INET6, address)
            return True
        except socket.error:  # invalid address
            return False
    else:  # do not know what to do? assume it is a valid address
        return True


def read_file(filename, mode='r'):
    # returns content from filename, making sure to close the file on exit.
    f = open(filename, mode)
    try:
        return f.read()
    finally:
        f.close()


def write_file(filename, value, mode='w'):
    # writes <value> to filename, making sure to close the file on exit.
    f = open(filename, mode)
    try:
        return f.write(value)
    finally:
        f.close()


def dict_to_sdict(obj):
    # convert dict and nested dicts to odict
    from .datastructures import sdict
    # if we have a dict, iterate over keys
    if isinstance(obj, dict) and not isinstance(obj, sdict):
        for k in obj:
            obj[k] = dict_to_sdict(obj[k])
        # convert the dict to convenient sdict object
        return sdict(obj)
    # return other objects without modifications
    return obj



from weppy import App, request, session, url, redirect, abort
from weppy.dal import DAL, Field, Model, belongs_to, has_many
from weppy.tools import requires
from weppy.tools.auth import Auth, AuthUser
from weppy.sessions import SessionCookieManager


app = App(__name__)


#: define models
class User(AuthUser):
    # will create "auth_user" table and groups/permissions ones
    has_many('posts', 'comments')


class Post(Model):
    belongs_to('user')
    has_many('comments')

    title = Field()
    text = Field('text')
    date = Field('datetime')

    default_values = {
        'user': lambda: session.auth.user.id,
        'date': lambda: request.now
    }
    validation = {
        'title': {'presence': True},
        'text': {'presence': True}
    }
    form_rw = {
        'user': False,
        'date': False
    }


class Comment(Model):
    belongs_to('user', 'post')

    text = Field('text')
    date = Field('datetime')

    default_values = {
        'user': lambda: session.auth.user.id,
        'date': lambda: request.now
    }
    validation = {
        'text': {'presence': True}
    }
    form_rw = {
        'user': False,
        'post': False,
        'date': False
    }

#: init db and auth
db = DAL(app)
auth = Auth(app, db, usermodel=User)
db.define_models(Post, Comment)


#: setup helping function
def setup_admin():
    # create the user
    user = User.create(
        email="walter@massivedynamics.com",
        first_name="Walter",
        last_name="Bishop",
        password="pocketuniverse"
    )
    # create an admin group
    admins = auth.add_group("admin")
    # add user to admins group
    auth.add_membership(admins, user.id)
    db.commit()


@app.command('setup')
def setup():
    setup_admin()

#: handlers
app.common_handlers = [
    SessionCookieManager('Walternate'),
    db.handler, auth.handler
]


#: exposing functions
@app.route("/")
def index():
    posts = Post.all().select(orderby=~Post.date)
    return dict(posts=posts)


@app.route("/post/<int:pid>")
def one(pid):
    def _validate_comment(form):
        # manually set post id in comment form
        form.params.post = pid
    # get post and return 404 if doesn't exist
    post = Post.get(pid)
    if not post:
        abort(404)
    # get comments and create a form for commenting
    comments = post.comments(orderby=~Comment.date)
    form = Comment.form(onvalidation=_validate_comment)
    if form.accepted:
        redirect(url('one', pid))
    return locals()


@app.route("/new")
@requires(lambda: auth.has_membership('admin'), url('index'))
def new_post():
    form = Post.form()
    if form.accepted:
        redirect(url('one', form.params.id))
    return dict(form=form)


@app.route('/account(/<str:f>)?(/<str:k>)?')
def account(f, k):
    form = auth(f, k)
    return dict(form=form)



# -*- coding: utf-8 -*-
"""
    weppy.dal.migrations.commands
    -----------------------------

    Provides command interfaces for migrations.

    :copyright: (c) 2014-2016 by Giovanni Barillari
    :license: BSD, see LICENSE for more details.
"""

from ...utils import cachedprop
from ...datastructures import sdict
from ..base import DAL
from .base import Schema
from .helpers import make_migration_id, to_tuple
from .operations import MigrationOp, UpgradeOps, DowngradeOps
from .scripts import ScriptDir


class Command(object):
    def __init__(self, app, dals):
        self.app = app
        self.dals = dals
        self.scriptdir = ScriptDir(self.app)

    def load_schema(self, db):
        schema_db_config = sdict(uri=db.config.uri)
        self.schema_db = DAL(
            self.app, schema_db_config, migrate=False, migrate_enabled=False)
        self.schema_db.define_models(Schema)
        self._ensure_schema_table_()
        self._load_current_revision_()

    def _ensure_schema_table_(self):
        # TODO -> has_table method in adapter, I don't like this "dirtness"
        try:
            self.schema_db(self.schema_db.Schema.id > 0).count()
        except:
            self.schema_db.rollback()
            q = self.schema_db._adapter.create_table(
                self.schema_db.Schema, migrate=False
            )
            self.schema_db._adapter.execute(q)
            self.schema_db.commit()

    def _load_current_revision_(self):
        revisions = self.schema_db(self.schema_db.Schema.id > 0).select()
        if not revisions:
            self._current_revision_ = []
        elif len(revisions) == 1:
            self._current_revision_ = [revisions[0].version]
        else:
            self._current_revision_ = [rev.version for rev in revisions]

    def _store_current_revision_(self, source, dest):
        logs = {
            'new': '> Adding revision %s to schema',
            'del': '> Removing revision %s from schema',
            'upd': '> Updating schema revision from %s to %s'}
        source = to_tuple(source)
        dest = to_tuple(dest)
        if source is None:
            print(logs['new'] % dest[0])
            self.schema_db.Schema.insert(version=dest[0])
            self.schema_db.commit()
            self._current_revision_ = [dest[0]]
            return
        if dest is None:
            print(logs['del'] % source[0])
            self.schema_db(self.schema_db.Schema.version == source[0]).delete()
            self.schema_db.commit()
            self._current_revision_ = []
            return
        if len(source) > 1:
            if len(source) > 2:
                self.schema_db(
                    self.schema_db.Schema.version.belongs(
                        source[1:])).delete()
                print(logs['del'] % source[1:])
            else:
                self.schema_db(
                    self.schema_db.Schema.version == source[1]).delete()
                print(logs['del'] % source[1])
            self.schema_db(self.schema_db.Schema.version == source[0]).update(
                version=dest[0]
            )
            print(logs['upd'] % (source[0], dest[0]))
            self._current_revision_ = [dest[0]]
        else:
            if list(source) != self._current_revision_:
                self.schema_db.Schema.insert(version=dest[0])
                print(logs['new'] % dest[0])
                self._current_revision_.append(dest[0])
            else:
                self.schema_db(
                    self.schema_db.Schema.version == source[0]
                ).update(
                    version=dest[0]
                )
                print(logs['upd'] % (source[0], dest[0]))
                self._current_revision_ = [dest[0]]
        self.schema_db.commit()

    @cachedprop
    def db(self):
        if len(self.dals) == 1:
            return self.dals[0]
        return self.dals

    def _generate_migration_script(self, migration, head):
        from .generation import Renderer
        upgrades, downgrades = Renderer.render_migration(migration)
        self.scriptdir.generate_revision(
            migration.rev_id, migration.message, head, upgrades=upgrades,
            downgrades=downgrades
        )

    def generate(self, message, head):
        if len(self.dals) > 1:
            raise RuntimeError('need just one db instance')
        from .generation import Generator
        upgrade_ops = Generator.generate_from(self.db, self.scriptdir, head)
        revid = make_migration_id()
        migration = MigrationOp(
            revid, upgrade_ops, upgrade_ops.reverse(), message)
        self._generate_migration_script(migration, head)
        print("> Generated migration for revision %s" % revid)

    def new(self, message, head):
        source_rev = self.scriptdir.get_revision(head)
        revid = make_migration_id()
        migration = MigrationOp(
            revid, UpgradeOps(), DowngradeOps(), message
        )
        self._generate_migration_script(migration, source_rev.revision)
        print("> Created new migration with revision %s" % revid)

    def history(self, base, head, verbose):
        print("> Migrations history")
        lines = []
        for sc in self.scriptdir.walk_revisions(
                base=base or "base",
                head=head or "heads"):
            lines.append(
                sc.cmd_format(
                    verbose=verbose, include_doc=True, include_parents=True))
        for line in lines:
            print(line)
        if not lines:
            print("No migrations for the selected application.")

    def status(self, verbose):
        self.load_schema(self.db)
        print("> Current revision(s) for %s" % self.db._uri)
        lines = []
        for rev in self.scriptdir.get_revisions(self._current_revision_):
            lines.append(rev.cmd_format(verbose))
        for line in lines:
            print(line)
        if not lines:
            print("No revision state found on the schema.")

    def up(self, rev_id):
        self.load_schema(self.db)
        start_point = self._current_revision_
        revisions = self.scriptdir.get_upgrade_revs(
            rev_id, start_point)
        print("> Performing upgrades against %s" % self.db._uri)
        for revision in revisions:
            print("> Performing upgrade: %s" % revision)
            migration = revision.migration_class(self.app, self.db)
            try:
                migration.up()
                self.db.commit()
                self._store_current_revision_(
                    migration.revises, migration.revision)
                print("> Succesfully upgraded to revision %s: %s" %
                      (revision.revision, revision.doc))
            except:
                self.db.rollback()
                print("> [ERROR] failed upgrading to %s" % revision)
                raise

    def down(self, rev_id):
        self.load_schema(self.db)
        start_point = self._current_revision_
        revisions = self.scriptdir.get_downgrade_revs(
            rev_id, start_point)
        print("> Performing downgrades against %s" % self.db._uri)
        for revision in revisions:
            print("> Performing downgrade: %s" % revision)
            migration = revision.migration_class(self.app, self.db)
            try:
                migration.down()
                self.db.commit()
                self._store_current_revision_(
                    migration.revision, migration.revises)
                print("> Succesfully downgraded from revision %s: %s" %
                      (revision.revision, revision.doc))
            except:
                self.db.rollback()
                print("> [ERROR] failed downgrading from %s" % revision)
                raise


def generate(app, dals, message, head):
    Command(app, dals).generate(message, head)


def new(app, dals, message, head):
    Command(app, dals).new(message, head)


def history(app, dals, rev_range, verbose):
    if rev_range is not None:
        if ":" not in rev_range:
            raise Exception(
                "History range requires [start]:[end], "
                "[start]:, or :[end]")
        base, head = rev_range.strip().split(":")
    else:
        base = head = None
    Command(app, dals).history(base, head, verbose)


def status(app, dals, verbose):
    Command(app, dals).status(verbose)


def up(app, dals, revision):
    Command(app, dals).up(revision)


def down(app, dals, revision):
    Command(app, dals).down(revision)



import re
from .._compat import to_unicode, to_native
from .basic import Validator
from .consist import isEmail
from .helpers import translate, options_sorter


# TODO port this
class isStrong(object):
    """
    enforces complexity requirements on a field
    """
    lowerset = frozenset(u'abcdefghijklmnopqrstuvwxyz')
    upperset = frozenset(u'ABCDEFGHIJKLMNOPQRSTUVWXYZ')
    numberset = frozenset(u'0123456789')
    sym1set = frozenset(u'!@#$%^&*()')
    sym2set = frozenset(u'~`-_=+[]{}\\|;:\'",.<>?/')
    otherset = frozenset(u'0123456789abcdefghijklmnopqrstuvwxyz')

    def __init__(self, min=None, max=None, upper=None, lower=None, number=None,
                 entropy=None,
                 special=None, specials=r'~!@#$%^&*()_+-=?<>,.:;{}[]|',
                 invalid=' "', error_message=None, es=False):
        self.entropy = entropy
        if entropy is None:
            # enforce default requirements
            self.min = 8 if min is None else min
            self.max = max  # was 20, but that doesn't make sense
            self.upper = 1 if upper is None else upper
            self.lower = 1 if lower is None else lower
            self.number = 1 if number is None else number
            self.special = 1 if special is None else special
        else:
            # by default, an entropy spec is exclusive
            self.min = min
            self.max = max
            self.upper = upper
            self.lower = lower
            self.number = number
            self.special = special
        self.specials = specials
        self.invalid = invalid
        self.error_message = error_message
        self.estring = es   # return error message as string (for doctest)

    def __call__(self, value):
        failures = []
        if value and len(value) == value.count('*') > 4:
            return (value, None)
        if self.entropy is not None:
            entropy = isStrong.calc_entropy(value)
            if entropy < self.entropy:
                failures.append(
                    translate(
                        "Entropy (%(have)s) less than required (%(need)s)")
                    % dict(have=entropy, need=self.entropy))
        if type(self.min) == int and self.min > 0:
            if not len(value) >= self.min:
                failures.append(translate("Minimum length is %s") % self.min)
        if type(self.max) == int and self.max > 0:
            if not len(value) <= self.max:
                failures.append(translate("Maximum length is %s") % self.max)
        if type(self.special) == int:
            all_special = [ch in value for ch in self.specials]
            if self.special > 0:
                if not all_special.count(True) >= self.special:
                    failures.append(
                        translate(
                            "Must include at least %s of the following: %s")
                        % (self.special, self.specials))
        if self.invalid:
            all_invalid = [ch in value for ch in self.invalid]
            if all_invalid.count(True) > 0:
                failures.append(
                    translate(
                        "May not contain any of the following: %s")
                    % self.invalid)
        if type(self.upper) == int:
            all_upper = re.findall("[A-Z]", value)
            if self.upper > 0:
                if not len(all_upper) >= self.upper:
                    failures.append(
                        translate("Must include at least %s upper case")
                        % str(self.upper))
            else:
                if len(all_upper) > 0:
                    failures.append(
                        translate("May not include any upper case letters"))
        if type(self.lower) == int:
            all_lower = re.findall("[a-z]", value)
            if self.lower > 0:
                if not len(all_lower) >= self.lower:
                    failures.append(
                        translate("Must include at least %s lower case")
                        % str(self.lower))
            else:
                if len(all_lower) > 0:
                    failures.append(
                        translate("May not include any lower case letters"))
        if type(self.number) == int:
            all_number = re.findall("[0-9]", value)
            if self.number > 0:
                numbers = "number"
                if self.number > 1:
                    numbers = "numbers"
                if not len(all_number) >= self.number:
                    failures.append(translate("Must include at least %s %s")
                                    % (str(self.number), numbers))
            else:
                if len(all_number) > 0:
                    failures.append(translate("May not include any numbers"))
        if len(failures) == 0:
            return (value, None)
        if not self.error_message:
            if self.estring:
                return (value, '|'.join(failures))
            from .templating import NOESCAPE
            return (value, NOESCAPE('<br />'.join(failures)))
        else:
            return (value, translate(self.error_message))

    @staticmethod
    def calc_entropy(string):
        " calculates a simple entropy for a given string "
        import math
        alphabet = 0    # alphabet size
        other = set()
        seen = set()
        lastset = None
        string = to_unicode(string)
        for c in string:
            # classify this character
            inset = isStrong.otherset
            for cset in (isStrong.lowerset, isStrong.upperset,
                         isStrong.numberset, isStrong.sym1set,
                         isStrong.sym2set):
                if c in cset:
                    inset = cset
                    break
            # calculate effect of character on alphabet size
            if inset not in seen:
                seen.add(inset)
                alphabet += len(inset)  # credit for a new character set
            elif c not in other:
                alphabet += 1   # credit for unique characters
                other.add(c)
            if inset is not lastset:
                alphabet += 1   # credit for set transitions
                lastset = cset
        entropy = len(
            string) * math.log(alphabet) / 0.6931471805599453  # math.log(2)
        return round(entropy, 2)


# TODO port this
class FilenameMatches(Validator):
    """
    Checks if name and extension of file uploaded through file input matches
    given criteria.

    Does *not* ensure the file type in any way. Returns validation failure
    if no data was uploaded.

    Args:
        filename: filename (before dot) regex
        extension: extension (after dot) regex
        lastdot: which dot should be used as a filename / extension separator:
            True means last dot, eg. file.png -> file / png
            False means first dot, eg. file.tar.gz -> file / tar.gz
        case:
            0 - keep the case,
            1 - transform the string into lowercase (default),
            2 - transform the string into uppercase

    If there is no dot present, extension checks will be done against empty
    string and filename checks against whole value.
    """

    def __init__(self, filename=None, extension=None, lastdot=True, case=1,
                 error_message='Enter valid filename'):
        if isinstance(filename, str):
            filename = re.compile(filename)
        if isinstance(extension, str):
            extension = re.compile(extension)
        self.filename = filename
        self.extension = extension
        self.lastdot = lastdot
        self.case = case
        self.error_message = error_message

    def __call__(self, value):
        try:
            string = value.filename
        except:
            return (value, translate(self.error_message))
        if self.case == 1:
            string = string.lower()
        elif self.case == 2:
            string = string.upper()
        if self.lastdot:
            dot = string.rfind('.')
        else:
            dot = string.find('.')
        if dot == -1:
            dot = len(string)
        if self.filename and not self.filename.match(string[:dot]):
            return (value, translate(self.error_message))
        elif self.extension and not self.extension.match(string[dot + 1:]):
            return (value, translate(self.error_message))
        else:
            return (value, None)


# Kept for reference (v0.3 and below)
class isEmailList(object):
    split_emails = re.compile('[^,;\s]+')

    def __init__(self, error_message='Invalid emails: %s'):
        self.error_message = error_message

    def __call__(self, value):
        bad_emails = []
        emails = []
        f = isEmail()
        for email in self.split_emails.findall(value):
            if email not in emails:
                emails.append(email)
            error = f(email)[1]
            if error and email not in bad_emails:
                bad_emails.append(email)
        if not bad_emails:
            return (value, None)
        else:
            return (value,
                    translate(self.error_message) % ', '.join(bad_emails))

    def formatter(self, value, row=None):
        return ', '.join(value or [])


# Kept for reference (v0.3 and below)
class inDb(Validator):
    """
    Used for reference fields, rendered as a dropbox
    """
    regex1 = re.compile('\w+\.\w+')
    regex2 = re.compile('%\(([^\)]+)\)\d*(?:\.\d+)?[a-zA-Z]')

    def __init__(
        self,
        dbset,
        field,
        label=None,
        error_message='Value not in database',
        orderby=None,
        groupby=None,
        distinct=None,
        cache=None,
        multiple=False,
        zero='',
        sort=False,
        _and=None,
    ):
        from pydal.objects import Table
        if isinstance(field, Table):
            field = field._id

        if hasattr(dbset, 'define_table'):
            self.dbset = dbset()
        else:
            self.dbset = dbset
        (ktable, kfield) = str(field).split('.')
        if not label:
            label = '%%(%s)s' % kfield
        if isinstance(label, str):
            if self.regex1.match(str(label)):
                label = '%%(%s)s' % str(label).split('.')[-1]
            ks = self.regex2.findall(label)
            if kfield not in ks:
                ks += [kfield]
            fields = ks
        else:
            ks = [kfield]
            fields = 'all'
        self.fields = fields
        self.label = label
        self.ktable = ktable
        self.kfield = kfield
        self.ks = ks
        self.error_message = error_message
        self.theset = None
        self.orderby = orderby
        self.groupby = groupby
        self.distinct = distinct
        self.cache = cache
        self.multiple = multiple
        self.zero = zero
        self.sort = sort
        self._and = _and

    def set_self_id(self, id):
        if self._and:
            self._and.record_id = id

    def build_set(self):
        from pydal.objects import FieldVirtual, FieldMethod
        table = self.dbset.db[self.ktable]
        if self.fields == 'all':
            fields = [f for f in table]
        else:
            fields = [table[k] for k in self.fields]
        ignore = (FieldVirtual, FieldMethod)
        fields = filter(lambda f: not isinstance(f, ignore), fields)
        if self.dbset.db._dbname != 'gae':
            orderby = self.orderby or reduce(lambda a, b: a | b, fields)
            groupby = self.groupby
            distinct = self.distinct
            dd = dict(orderby=orderby, groupby=groupby,
                      distinct=distinct, cache=self.cache,
                      cacheable=True)
            records = self.dbset(table).select(*fields, **dd)
        else:
            orderby = self.orderby or \
                reduce(lambda a, b: a | b, (
                    f for f in fields if not f.name == 'id'))
            dd = dict(orderby=orderby, cache=self.cache, cacheable=True)
            records = self.dbset(table).select(table.ALL, **dd)
        self.theset = [str(r[self.kfield]) for r in records]
        if isinstance(self.label, str):
            self.labels = [self.label % r for r in records]
        else:
            self.labels = [self.label(r) for r in records]

    def options(self, zero=True):
        self.build_set()
        items = [(k, self.labels[i]) for (i, k) in enumerate(self.theset)]
        if self.sort:
            items.sort(options_sorter)
        if zero and self.zero is not None and not self.multiple:
            items.insert(0, ('', self.zero))
        return items

    def __call__(self, value):
        table = self.dbset.db[self.ktable]
        field = table[self.kfield]
        if self.multiple:
            if self._and:
                raise NotImplementedError
            if isinstance(value, list):
                values = value
            elif value:
                values = [value]
            else:
                values = []
            if isinstance(self.multiple, (tuple, list)) and \
                    not self.multiple[0] <= len(values) < self.multiple[1]:
                return values, translate(self.error_message)
            if self.theset:
                if not [v for v in values if v not in self.theset]:
                    return values, None
            else:
                from pydal.adapters import GoogleDatastoreAdapter

                def count(values, s=self.dbset, f=field):
                    return s(f.belongs(map(int, values))).count()
                if isinstance(self.dbset.db._adapter, GoogleDatastoreAdapter):
                    range_ids = range(0, len(values), 30)
                    total = sum(count(values[i:i + 30]) for i in range_ids)
                    if total == len(values):
                        return values, None
                elif count(values) == len(values):
                    return values, None
        elif self.theset:
            if str(value) in self.theset:
                if self._and:
                    return self._and(value)
                else:
                    return value, None
        else:
            if self.dbset(field == value).count():
                if self._and:
                    return self._and(value)
                else:
                    return value, None
        return value, translate(self.error_message)


# Kept for reference (v0.3 and below)
class notInDb(Validator):
    """
    makes the field unique
    """

    def __init__(
        self,
        dbset,
        field,
        error_message='Value already in database or empty',
        allowed_override=[],
        ignore_common_filters=False,
    ):

        from pydal.objects import Table
        if isinstance(field, Table):
            field = field._id

        if hasattr(dbset, 'define_table'):
            self.dbset = dbset()
        else:
            self.dbset = dbset
        self.field = field
        self.error_message = error_message
        self.record_id = 0
        self.allowed_override = allowed_override
        self.ignore_common_filters = ignore_common_filters

    def set_self_id(self, id):
        self.record_id = id

    def __call__(self, value):
        value = to_native(to_unicode(value))
        if not value.strip():
            return value, translate(self.error_message)
        if value in self.allowed_override:
            return value, None
        (tablename, fieldname) = str(self.field).split('.')
        table = self.dbset.db[tablename]
        field = table[fieldname]
        subset = self.dbset(field == value,
                            ignore_common_filters=self.ignore_common_filters)
        id = self.record_id
        if isinstance(id, dict):
            fields = [table[f] for f in id]
            row = subset.select(*fields, **dict(
                limitby=(0, 1), orderby_on_limitby=False)).first()
            if row and any(str(row[f]) != str(id[f]) for f in id):
                return value, translate(self.error_message)
        else:
            row = subset.select(table._id, field, limitby=(0, 1),
                                orderby_on_limitby=False).first()
            if row and str(row.id) != str(id):
                return value, translate(self.error_message)
        return value, None



# -*- coding: utf-8 -*-
"""
    weppy.templating.cache
    ----------------------

    Provides cache for templating system.

    :copyright: (c) 2014-2016 by Giovanni Barillari
    :license: BSD, see LICENSE for more details.
"""

import hashlib
from .._compat import iteritems, to_bytes
from ..utils import cachedprop


def make_md5(value):
    return hashlib.md5(to_bytes(value)).hexdigest()[:8]


class TemplaterCache(object):
    data = {}
    pdata = {}
    hashes = {}
    dependencies = {}

    def __init__(self, app, templater):
        self.app = app
        self.templater = templater
        self.tpath = app.template_path

    @cachedprop
    def changes(self):
        return self.app.debug or self.app.config.templates_auto_reload

    @cachedprop
    def get(self):
        if self.changes:
            return self.reloader_get
        return self.cached_get

    def _fetch_dependency_source(self, filename):
        tpath, tname = self.templater.preload(self.tpath, filename)
        tsource = self.templater.load(tpath, tname)
        return self.templater.prerender(tsource, tname)

    def reloader_get(self, filename, source):
        hashed = make_md5(source)
        if self.hashes.get(filename) != hashed:
            return None, None
        for iname, ihash in iteritems(self.dependencies[filename]):
            hashed = make_md5(self._fetch_dependency_source(iname))
            if ihash != hashed:
                return None, None
        return self.cached_get(filename, source)

    def cached_get(self, filename, source):
        return self.data.get(filename), self.pdata.get(filename)

    def set(self, filename, source, compiled, pdata, included):
        self.data[filename] = compiled
        self.pdata[filename] = pdata
        if self.changes:
            self.hashes[filename] = make_md5(source)
            self.dependencies[filename] = {}
            for iname, isource in included:
                self.dependencies[filename][iname] = make_md5(isource)



# -*- coding: utf-8 -*-
"""
    weppy.templating.core
    ---------------------

    Provides the templating system for weppy.

    :copyright: (c) 2014-2016 by Giovanni Barillari
    :license: BSD, see LICENSE for more details.
"""

import os
import cgi
import sys
from .._compat import StringIO, reduce, string_types, text_type, to_native, \
    to_unicode, to_bytes
from ..tags import asis
from ..helpers import load_component
from ..datastructures import sdict
from .parser import TemplateParser
from .cache import TemplaterCache
from .helpers import TemplateMissingError, TemplateReference


class DummyResponse():
    def __init__(self):
        self.body = StringIO()

    @staticmethod
    def _to_html(data):
        return to_bytes(data, 'ascii', 'xmlcharrefreplace')

    @staticmethod
    def _to_native(data):
        if not isinstance(data, text_type):
            data = to_unicode(data)
        return to_native(data)

    @staticmethod
    def _to_unicode(data):
        if not isinstance(data, string_types):
            return text_type(data)
        return to_unicode(data)

    def write(self, data, escape=True):
        body = None
        if escape:
            if hasattr(data, 'to_html'):
                try:
                    body = to_native(data.to_html())
                except:
                    pass
            if body is None:
                body = self._to_native(
                    self._to_html(
                        cgi.escape(
                            self._to_unicode(data), True
                        ).replace(u"'", u"&#x27;")))
        else:
            body = self._to_native(data)
        self.body.write(body)


class Templater(object):
    def __init__(self, application):
        self.loaders = application.template_preloaders
        self.renders = application.template_extensions
        self.lexers = application.template_lexers
        self.cache = TemplaterCache(application, self)

    def preload(self, path, name):
        fext = os.path.splitext(name)[1]
        return reduce(lambda s, e: e.preload(s[0], s[1]),
                      self.loaders.get(fext, []), (path, name))

    def load(self, path, filename):
        # return source as unicode str
        try:
            file_obj = open(os.path.join(path, filename), 'r')
            source = to_unicode(file_obj.read())
            file_obj.close()
        except IOError:
            raise TemplateMissingError(path, filename)
        return source

    def prerender(self, source, filename):
        return reduce(lambda s, e: e.preprocess(s, filename),
                      self.renders, source)

    def inject(self, context):
        for extension in self.renders:
            extension.inject(context)

    def parse(self, path, filename, source, context):
        code, parserdata = self.cache.get(filename, source)
        if not code:
            parser = TemplateParser(self, source, name=filename,
                                    context=context, path=path)
            code = compile(str(parser), filename, 'exec')
            parserdata = sdict(content=parser.content,
                               blocks=parser.content.blocks)
            self.cache.set(
                filename, source, code, parserdata, parser.included_templates)
        return code, parserdata

    def _render(self, source='', path=None, filename=None, context={}):
        if isinstance(context, sdict):
            context = dict(context)
        context['asis'] = context.get('asis', asis)
        context['load_component'] = context.get(
            'load_component', load_component)
        context['_DummyResponse_'] = DummyResponse()
        code, parserdata = self.parse(path, filename, source, context)
        self.inject(context)
        try:
            exec(code, context)
        except:
            from ..debug import make_traceback
            exc_info = sys.exc_info()
            try:
                parserdata.path = path
                parserdata.name = filename
                template_ref = TemplateReference(parserdata, code, exc_info[0],
                                                 exc_info[1], exc_info[2])
            except:
                template_ref = None
            context['__weppy_template__'] = template_ref
            make_traceback(exc_info, template_ref)
        return context['_DummyResponse_'].body.getvalue()

    def render(self, path, filename, context={}):
        tpath, tname = self.preload(path, filename)
        tsource = self.load(tpath, tname)
        tsource = self.prerender(tsource, tname)
        return self._render(tsource, tpath, tname, context)



# -*- coding: utf-8 -*-
"""
    weppy.templating.parser
    -----------------------

    Provides the templating parser.

    :copyright: (c) 2014-2016 by Giovanni Barillari

    Based on the web2py's templating system (http://www.web2py.com)
    :copyright: (c) by Massimo Di Pierro <mdipierro@cs.depaul.edu>

    :license: BSD, see LICENSE for more details.
"""

from re import compile, sub, escape, DOTALL
from .._compat import implements_to_string, to_unicode
from .contents import Node, BlockNode, Content
from .helpers import TemplateError
from .lexers import default_lexers


@implements_to_string
class TemplateParser(object):
    default_delimiters = ('{{', '}}')
    r_tag = compile(r'(\{\{.*?\}\})', DOTALL)

    r_multiline = compile(r'(""".*?""")|(\'\'\'.*?\'\'\')', DOTALL)

    # These are used for re-indentation.
    # Indent + 1
    re_block = compile('^(elif |else:|except:|except |finally:).*$', DOTALL)

    # Indent - 1
    re_unblock = compile('^(return|continue|break|raise)( .*)?$', DOTALL)
    # Indent - 1
    re_pass = compile('^pass( .*)?$', DOTALL)

    def __init__(self, templater, text, name="ParserContainer", context={},
                 path='templates/', writer='_DummyResponse_.write', lexers={},
                 delimiters=('{{', '}}'), _super_nodes=[]):
        """
        text -- text to parse
        context -- context to parse in
        path -- folder path to templates
        writer -- string of writer class to use
        lexers -- dict of custom lexers to use.
        delimiters -- for example ('{{','}}')
        _super_nodes -- a list of nodes to check for inclusion
                        this should only be set by "self.extend"
                        It contains a list of SuperNodes from a child
                        template that need to be handled.
        """
        self.templater = templater
        # Keep a root level name.
        self.name = name
        # Raw text to start parsing.
        self.text = text
        # Writer to use (refer to the default for an example).
        # This will end up as
        # "%s(%s, escape=False)" % (self.writer, value)
        self.writer = writer

        # Dictionary of custom name lexers to use.
        #if isinstance(lexers, dict):
        #    self.lexers = lexers
        #else:
        #    self.lexers = {}
        self.lexers = default_lexers
        self.lexers.update(self.templater.lexers)

        # Path of templates
        self.path = path
        # Context for templates.
        self.context = context

        # allow optional alternative delimiters
        self.delimiters = delimiters
        if delimiters != self.default_delimiters:
            escaped_delimiters = (escape(delimiters[0]),
                                  escape(delimiters[1]))
            self.r_tag = compile(r'(%s.*?%s)' % escaped_delimiters, DOTALL)
        #elif hasattr(context.get('response', None), 'delimiters'):
        #    if context['response'].delimiters != self.default_delimiters:
        #        escaped_delimiters = (
        #            escape(context['response'].delimiters[0]),
        #            escape(context['response'].delimiters[1]))
        #        self.r_tag = compile(r'(%s.*?%s)' % escaped_delimiters,
        #                             DOTALL)

        # Create a root level Content that everything will go into.
        self.content = Content(name=name)

        # Stack will hold our current stack of nodes.
        # As we descend into a node, it will be added to the stack
        # And when we leave, it will be removed from the stack.
        # self.content should stay on the stack at all times.
        self.stack = [self.content]

        # This variable will hold a reference to every super block
        # that we come across in this template.
        self.super_nodes = []

        # This variable will hold a reference to the child
        # super nodes that need handling.
        self.child_super_nodes = _super_nodes

        # This variable will hold a reference to every block
        # that we come across in this template
        self.blocks = {}

        self.included_templates = []
        self.current_lines = (1, 1)

        # Begin parsing.
        self.parse(text)

    def create_block(self, name=None, pre_extend=False, delimiters=None):
        return BlockNode(name=name, pre_extend=pre_extend,
                         delimiters=delimiters or self.delimiters)

    def create_node(self, value, pre_extend=False, use_writer=True,
                    writer_escape=True):
        if use_writer:
            if not writer_escape:
                value = u"\n%s(%s, escape=False)" % (self.writer, value)
            else:
                value = u"\n%s(%s)" % (self.writer, value)
        else:
            value = u"\n%s" % value
        return Node(value, pre_extend=pre_extend, template=self.name,
                    lines=self.current_lines)

    def create_htmlnode(self, value, pre_extend=False):
        value = u"\n%s(%r, escape=False)" % (self.writer, value)
        return Node(value, pre_extend=pre_extend, template=self.name,
                    lines=self.current_lines)

    def __str__(self):
        return self.reindent(to_unicode(self.content))

    def reindent(self, text):
        """
        Reindents a string of unindented python code.
        """

        # Get each of our lines into an array.
        lines = text.split('\n')

        # Our new lines
        new_lines = []

        # Keeps track of how many indents we have.
        # Used for when we need to drop a level of indentation
        # only to reindent on the next line.
        credit = 0

        # Current indentation
        k = 0

        #################
        # THINGS TO KNOW
        #################

        # k += 1 means indent
        # k -= 1 means unindent
        # credit = 1 means unindent on the next line.

        for raw_line in lines:
            line = raw_line.strip()

            # ignore empty lines
            if not line:
                continue

            # If we have a line that contains python code that
            # should be unindented for this line of code.
            # and then reindented for the next line.
            if TemplateParser.re_block.match(line):
                k = k + credit - 1

            # We obviously can't have a negative indentation
            k = max(k, 0)

            # Add the indentation!
            new_lines.append(u' ' * (4 * k) + line)

            # Bank account back to 0 again :(
            credit = 0

            # If we are a pass block, we obviously de-dent.
            if TemplateParser.re_pass.match(line):
                k -= 1

            # If we are any of the following, de-dent.
            # However, we should stay on the same level
            # But the line right after us will be de-dented.
            # So we add one credit to keep us at the level
            # while moving back one indentation level.
            if TemplateParser.re_unblock.match(line):
                credit = 1
                k -= 1

            # If we are an if statement, a try, or a semi-colon we
            # probably need to indent the next line.
            if line.endswith(':') and not line.startswith('#'):
                k += 1

        # This must come before so that we can raise an error with the
        # right content.
        new_text = u'\n'.join(new_lines)

        if k > 0:
            #self._raise_error('missing "pass" in view', new_text)
            raise TemplateError(self.path, 'missing "pass" in view',
                                self.name, 1)
        elif k < 0:
            #self._raise_error('too many "pass" in view', new_text)
            raise TemplateError(self.path, 'too many "pass" in view',
                                self.name, 1)

        return new_text

    def _get_file_text(self, filename):
        """
        Attempt to open ``filename`` and retrieve its text.

        This will use self.path to search for the file.
        """

        # If they didn't specify a filename, how can we find one!
        if not filename.strip():
            #self._raise_error('Invalid template filename')
            raise TemplateError(self.path, 'Invalid template filename',
                                self.name, 1)

        # Allow Views to include other views dynamically
        context = self.context
        #if current and not "response" in context:
        #    context["response"] = getattr(current, 'response', None)

        # Get the filename; filename looks like ``"template.html"``.
        # We need to eval to remove the quotes and get the string type.
        filename = eval(filename, context)

        # Get the path of the file on the system.
        #filepath = self.path and os.path.join(self.path, filename) or filename

        # Get the file and read the content
        tpath, tname = self.templater.preload(
            self.path, filename)
        try:
            tsource = self.templater.load(tpath, tname)
        except:
            raise TemplateError(self.path, 'Unable to open included view file',
                                self.name, 1)
        tsource = self.templater.prerender(tsource, tname)

        self.included_templates.append((tname, tsource))

        return tsource, tname

    def include(self, content, filename):
        """
        Include ``filename`` here.
        """
        text, tname = self._get_file_text(filename)

        t = TemplateParser(self.templater, text,
                           name=tname,
                           context=self.context,
                           path=self.path,
                           writer=self.writer,
                           delimiters=self.delimiters)

        content.append(t.content)

    def extend(self, filename):
        """
        Extend ``filename``. Anything not declared in a block defined by the
        parent will be placed in the parent templates ``{{include}}`` block.
        """
        text, tname = self._get_file_text(filename)

        # Create out nodes list to send to the parent
        super_nodes = []
        # We want to include any non-handled nodes.
        super_nodes.extend(self.child_super_nodes)
        # And our nodes as well.
        super_nodes.extend(self.super_nodes)

        t = TemplateParser(self.templater, text,
                           name=tname,
                           context=self.context,
                           path=self.path,
                           writer=self.writer,
                           delimiters=self.delimiters,
                           _super_nodes=super_nodes)

        # Make a temporary buffer that is unique for parent
        # template.
        buf = BlockNode(
            name='__include__' + tname, delimiters=self.delimiters)
        pre = []

        # Iterate through each of our nodes
        for node in self.content.nodes:
            # If a node is a block
            if isinstance(node, BlockNode):
                # That happens to be in the parent template
                if node.name in t.content.blocks:
                    # Do not include it
                    continue

            if isinstance(node, Node):
                # Or if the node was before the extension
                # we should not include it
                if node.pre_extend:
                    pre.append(node)
                    continue

            # Otherwise, it should go int the
            # Parent templates {{include}} section.
                buf.append(node)
            else:
                buf.append(node)

        # Clear our current nodes. We will be replacing this with
        # the parent nodes.
        self.content.nodes = []

        t_content = t.content

        # Set our include, unique by filename
        t_content.blocks['__include__' + tname] = buf

        # Make sure our pre_extended nodes go first
        t_content.insert(pre)

        # Then we extend our blocks
        t_content.extend(self.content)

        # Work off the parent node.
        self.content = t_content

    def parse(self, text):
        # Basically, r_tag.split will split the text into
        # an array containing, 'non-tag', 'tag', 'non-tag', 'tag'
        # so if we alternate this variable, we know
        # what to look for. This is alternate to
        # line.startswith("{{")
        self._in_tag = False
        self._needs_extend = None
        self._is_pre_extend = True
        last_was_code_block = False

        # Use a list to store everything in
        # This is because later the code will "look ahead"
        # for missing strings or brackets.
        ij = self.r_tag.split(text)
        # j = current index
        # i = current item

        stack = self.stack
        for j in range(len(ij)):
            i = ij[j]

            if i:
                self.current_lines = (self.current_lines[1],
                                      self.current_lines[1] +
                                      len(i.split("\n")) - 1)

                if not stack:
                    raise TemplateError(
                        self.path, 'The "end" tag is unmatched, please check' +
                        ' if you have a starting "block" tag', self.name,
                        self.current_lines)

                # Our current element in the stack.
                top = stack[-1]

                if self._in_tag:
                    line = i

                    # Get rid of '{{' and '}}'
                    line = line[2:-2].strip()

                    # This is bad juju, but let's do it anyway
                    if not line:
                        continue

                    # We do not want to replace the newlines in code,
                    # only in block comments.
                    def remove_newline(re_val):
                        # Take the entire match and replace newlines with
                        # escaped newlines.
                        return re_val.group(0).replace('\n', '\\n')

                    # Perform block comment escaping.
                    # This performs escaping ON anything
                    # in between """ and """
                    line = sub(TemplateParser.r_multiline,
                               remove_newline,
                               line)

                    if line.startswith('='):
                        # IE: {{=response.title}}
                        name, value = '=', line[1:].strip()
                    else:
                        v = line.split(' ', 1)
                        if len(v) == 1:
                            # Example
                            # {{ include }}
                            # {{ end }}
                            name = v[0]
                            value = u''
                        else:
                            # Example
                            # {{ block pie }}
                            # {{ include "layout.html" }}
                            # {{ for i in range(10): }}
                            name = v[0]
                            value = v[1]

                    # This will replace newlines in block comments
                    # with the newline character. This is so that they
                    # retain their formatting, but squish down to one
                    # line in the rendered template.

                    # Check presence of an appropriate lexer
                    lexer = self.lexers.get(name)
                    if lexer and not value.startswith('='):
                        lexer(parser=self, value=value)
                        last_was_code_block = False
                    else:
                        # If we don't know where it belongs
                        # we just add it anyways without formatting.
                        if line and self._in_tag:

                            # Split on the newlines >.<
                            tokens = line.split('\n')

                            # We need to look for any instances of
                            # for i in range(10):
                            #   = i
                            # pass
                            # So we can properly put a writer in place.
                            continuation = False
                            len_parsed = 0
                            for k, token in enumerate(tokens):

                                token = tokens[k] = token.strip()
                                len_parsed += len(token)

                                if token.startswith('='):
                                    if token.endswith('\\'):
                                        continuation = True
                                        tokens[k] = u"%s(%s" % (
                                            self.writer, token[1:].strip())
                                    else:
                                        tokens[k] = u"%s(%s)" % (
                                            self.writer, token[1:].strip())
                                elif continuation:
                                    tokens[k] += u')'
                                    continuation = False

                            buf = u'\n'.join(tokens)
                            node = self.create_node(buf, self._is_pre_extend,
                                                    use_writer=False)
                            top.append(node)
                            last_was_code_block = True

                else:
                    if not last_was_code_block or (
                            last_was_code_block and i.strip()):
                        # It is HTML so just include it.
                        node = self.create_htmlnode(i, self._is_pre_extend)
                        top.append(node)
                    last_was_code_block = False

            # Remember: tag, not tag, tag, not tag
            self._in_tag = not self._in_tag

        # Make a list of items to remove from child
        to_rm = []

        # Go through each of the children nodes
        for node in self.child_super_nodes:
            # If we declared a block that this node wants to include
            if node.name in self.blocks:
                # Go ahead and include it!
                node.value = self.blocks[node.name]
                # Since we processed this child, we don't need to
                # pass it along to the parent
                to_rm.append(node)

        # Remove some of the processed nodes
        for node in to_rm:
            # Since this is a pointer, it works beautifully.
            # Sometimes I miss C-Style pointers... I want my asterisk...
            self.child_super_nodes.remove(node)

        # If we need to extend a template.
        if self._needs_extend:
            self.extend(self._needs_extend)



import pytest
from weppy import response, session
from bloggy import app, db, setup_admin


@pytest.fixture()
def client():
    return app.test_client()


@pytest.fixture(scope='module')
def logged_client():
    db._adapter.reconnect()
    setup_admin()
    c = app.test_client()
    with c.get('/account/login').context as ctx:
        c.post('/account/login', data={
            'email': 'walter@massivedynamics.com',
            'password': 'pocketuniverse',
            '_csrf_token': list(ctx.session._csrf)[-1]
        }, follow_redirects=True)
        return c


def test_empty_db(client):
    rv = client.get('/')
    assert 'No posts here so far' in rv.data


def test_login(logged_client):
    logged_client.get('/')
    assert session.auth.user is not None


def test_no_admin_access(client):
    client.get('/new')
    assert response.status == 303


def test_admin_access(logged_client):
    logged_client.get('/new')
    assert response.status == 200



"""
    weppy.handlers
    --------------

    Provide the Handler class and wrapper to process handlers.

    :copyright: (c) 2014-2016 by Giovanni Barillari
    :license: BSD, see LICENSE for more details.
"""

from ._compat import basestring
from .http import HTTP


class _wrapWithHandlers(object):
    def __init__(self, handlers=[]):
        self.handlers = handlers

    def __call__(self, f):
        def wrap(f, handler):
            def handled_function(*args, **kwargs):
                try:
                    handler.on_start()
                    output = handler.wrap_call(f)(*args, **kwargs)
                    handler.on_success()
                    return output
                except HTTP:
                    handler.on_success()
                    raise
                except:
                    handler.on_failure()
                    raise
            return handled_function
        for handler in reversed(self.handlers):
            if isinstance(handler, Handler):
                f = wrap(f, handler)
        return f


class Handler(object):
    def on_start(self):
        pass

    def on_success(self):
        pass

    def on_failure(self):
        pass

    def wrap_call(self, func):
        return func

    def on_end(self):
        pass


class RequireHandler(Handler):
    def __init__(self, condition=None, otherwise=None):
        if condition is None or otherwise is None:
            raise SyntaxError('usage: @requires(condition, otherwise)')
        if not callable(otherwise) and not isinstance(otherwise, basestring):
            raise SyntaxError("'otherwise' param must be string or callable")
        self.condition = condition
        self.otherwise = otherwise

    def wrap_call(self, func):
        def wrap(*args, **kwargs):
            flag = self.condition()
            if not flag:
                from .http import redirect
                if self.otherwise is not None:
                    if callable(self.otherwise):
                        return self.otherwise()
                    redirect(self.otherwise)
                else:
                    from .helpers import flash
                    flash('Insufficient privileges')
                    redirect('/')
            return func(*args, **kwargs)
        return wrap


class Helper(Handler):
    def _inject(self, d):
        exclude = ['on_start', 'on_end', 'on_success', 'on_failure',
                   'wrap_call']
        attr_list = dir(self)
        for attr in attr_list:
            if attr.startswith('_') or attr in exclude:
                continue
            d[attr] = self.__getattribute__(attr)
        return d

    def wrap_call(self, func):
        def g(*a, **b):
            output = func(*a, **b)
            if isinstance(output, dict):
                output = self._inject(output)
            return output
        return g



# -*- coding: utf-8 -*-
"""
    weppy.http
    ----------

    Provides the HTTP interface for weppy.

    :copyright: (c) 2014-2016 by Giovanni Barillari
    :license: BSD, see LICENSE for more details.
"""

from ._compat import text_type, to_bytes


class HTTP(Exception):
    status_codes = {
        200: 'OK',
        201: 'CREATED',
        202: 'ACCEPTED',
        203: 'NON-AUTHORITATIVE INFORMATION',
        204: 'NO CONTENT',
        205: 'RESET CONTENT',
        206: 'PARTIAL CONTENT',
        301: 'MOVED PERMANENTLY',
        302: 'FOUND',
        303: 'SEE OTHER',
        304: 'NOT MODIFIED',
        305: 'USE PROXY',
        307: 'TEMPORARY REDIRECT',
        400: 'BAD REQUEST',
        401: 'UNAUTHORIZED',
        403: 'FORBIDDEN',
        404: 'NOT FOUND',
        405: 'METHOD NOT ALLOWED',
        406: 'NOT ACCEPTABLE',
        407: 'PROXY AUTHENTICATION REQUIRED',
        408: 'REQUEST TIMEOUT',
        409: 'CONFLICT',
        410: 'GONE',
        411: 'LENGTH REQUIRED',
        412: 'PRECONDITION FAILED',
        413: 'REQUEST ENTITY TOO LARGE',
        414: 'REQUEST-URI TOO LONG',
        415: 'UNSUPPORTED MEDIA TYPE',
        416: 'REQUESTED RANGE NOT SATISFIABLE',
        417: 'EXPECTATION FAILED',
        422: 'UNPROCESSABLE ENTITY',
        500: 'INTERNAL SERVER ERROR',
        501: 'NOT IMPLEMENTED',
        502: 'BAD GATEWAY',
        503: 'SERVICE UNAVAILABLE',
        504: 'GATEWAY TIMEOUT',
        505: 'HTTP VERSION NOT SUPPORTED',
    }

    def __init__(self, status_code, body=u'', headers=None, cookies=None):
        self.status_code = status_code
        self.status_name = HTTP.status_codes.get(status_code, status_code)
        self._set_body(body)
        self.headers = list(headers.items()) if headers else []
        if cookies:
            self.headers += HTTP.cookies2header(cookies)
        #if status_code != 200:
        #    if str(status_code)[0] == '4' and len(self.body) < 512:
        #        self.body += '<!-- %s //-->' % ('x' * 512)  # trick IE

    def _set_body(self, body):
        if body is None:
            self.body = []
        elif isinstance(body, (text_type, bytes, bytearray)):
            body = to_bytes(body)
            self.body = [body]
        else:
            self.body = body

    def to(self, environ, start_response):
        start_response(
            "%s %s" % (self.status_code, self.status_name), self.headers)
        if environ.get('REQUEST_METHOD', '') == 'HEAD':
            return [to_bytes('')]
        return self.body

    @staticmethod
    def cookies2header(cookies):
        headers = []
        for cookie in cookies.values():
            headers.append(('Set-Cookie', str(cookie)[11:]))
        return headers

    @staticmethod
    def redirect(location, status_code=303):
        from .globals import current
        current.response.status = status_code
        location = location.replace('\r', '%0D').replace('\n', '%0A')
        raise HTTP(status_code, headers=dict(Location=location))


redirect = HTTP.redirect



# -*- coding: utf-8 -*-
"""
    weppy.templating.lexers
    -----------------------

    Provides the default lexers for templating parsing
    (using the same logic applied for extensions).

    :copyright: (c) 2014-2016 by Giovanni Barillari
    :license: BSD, see LICENSE for more details.
"""

from .._compat import to_unicode
from ..extensions import TemplateLexer
from .contents import SuperNode, BlockNode


class WeppyLexer(TemplateLexer):
    evaluate_value = False

    def __init__(self):
        pass


class DefineLexer(WeppyLexer):
    def process(self, value):
        #: insert a variable in the template
        node = self.parser.create_node(value, self.parser._is_pre_extend)
        self.top.append(node)


class BlockLexer(WeppyLexer):
    def process(self, value):
        #: create a new node with name
        node = self.parser.create_block(
            value.strip(), self.parser._is_pre_extend)
        #: append this node to the active one
        self.top.append(node)
        #: add the node to the stack so anything after this gets added
        #  to this node. This allows us to "nest" nodes.
        self.stack.append(node)


class EndLexer(WeppyLexer):
    def process(self, value):
        #: we are done with this node, let's store the instance
        self.parser.blocks[self.top.name] = self.top
        #: and pop it
        self.stack.pop()


class SuperLexer(WeppyLexer):
    def process(self, value):
        #: get correct target name, if not provided assume the top block
        target_node = value if value else self.top.name
        #: create a SuperNode instance
        node = SuperNode(name=target_node,
                         pre_extend=self.parser._is_pre_extend)
        #: add new node to the super_nodes list
        self.parser.super_nodes.append(node)
        #: put the node into the tree
        self.top.append(node)


class IncludeLexer(WeppyLexer):
    def process(self, value):
        #: if we have a value, just call the parser function
        if value:
            self.parser.include(self.top, value)
        #: otherwise, make a temporary include node that the child node
        #  will know to hook into.
        else:
            include_node = BlockNode(
                name='__include__' + self.parser.name,
                pre_extend=self.parser._is_pre_extend,
                delimiters=self.parser.delimiters)
            self.top.append(include_node)


class ExtendLexer(WeppyLexer):
    def process(self, value):
        #: extend the proper template
        self.parser._needs_extend = value
        self.parser._is_pre_extend = False


class HelpersLexer(WeppyLexer):
    helpers = [
        '<script type="text/javascript" ' +
        'src="/__weppy__/jquery.min.js"></script>',
        '<script type="text/javascript" ' +
        'src="/__weppy__/helpers.js"></script>']

    def process(self, value):
        node = self.parser.create_htmlnode(
            u"\n".join(h for h in self.helpers), self.parser._is_pre_extend)
        self.top.append(node)


class MetaLexer(WeppyLexer):
    def process(self, value):
        if not value:
            value = u'current.response.get_meta()'
        node = self.parser.create_node(
            value, self.parser._is_pre_extend, writer_escape=False)
        self.top.append(node)


class StaticLexer(WeppyLexer):
    evaluate_value = True

    def process(self, value):
        from ..expose import url
        file_name = value.split("?")[0]
        surl = to_unicode(url('static', file_name))
        file_ext = file_name.rsplit(".", 1)[-1]
        if file_ext == 'js':
            s = u'<script type="text/javascript" src="%s"></script>' % surl
        elif file_ext == "css":
            s = u'<link rel="stylesheet" href="%s" type="text/css" />' % surl
        else:
            s = None
        if s:
            node = self.parser.create_htmlnode(s, self.parser._is_pre_extend)
            self.top.append(node)


default_lexers = {
    '=': DefineLexer(),
    'block': BlockLexer(),
    'end': EndLexer(),
    'super': SuperLexer(),
    'include': IncludeLexer(),
    'extend': ExtendLexer(),
    'include_helpers': HelpersLexer(),
    'include_meta': MetaLexer(),
    'include_static': StaticLexer()
}



#!/usr/bin/env python
# -*- coding: utf8 -*-
# Plural-Forms for en (English)

# English language has 2 forms: 1 singular and 1 plural
nplurals = 2


# Determine plural_id for number *n* as sequence of positive
# integers: 0,1,...
# NOTE! For singular form ALWAYS return plural_id = 0
get_plural_id = lambda n: int(n != 1)

# Construct and return plural form of *word* using
# *plural_id* (which ALWAYS>0). This function will be executed
# for words (or phrases) not found in plural_dict dictionary
construct_plural_form = lambda word, plural_id: (
                        word +  # NOQA
                            ('es'  # NOQA
                                if word[-1:] in ('s', 'x', 'o') or
                                word[-2:] in ('sh', 'ch')
                                else 's')
                        )  # NOQA



# -*- coding: utf-8 -*-
"""
    weppy.forms
    -----------

    Provides classes to create and style forms in weppy.

    :copyright: (c) 2014-2016 by Giovanni Barillari
    :license: BSD, see LICENSE for more details.
"""

from ._compat import iteritems, iterkeys
from .dal import Field
from .datastructures import sdict
from .globals import current, request, session
from .security import CSRFStorage
from .tags import TAG, tag, cat, asis
from .utils import cachedprop

__all__ = ['Form', 'DALForm']


class Form(TAG):
    default_attrs = {
        '_action': '', '_method': 'POST', '_enctype': 'multipart/form-data',
        'submit': 'Submit', 'csrf': 'auto', 'keepvalues': False,
        'onvalidation': None, 'id_prefix': '', 'upload': None
    }

    @staticmethod
    def _get_default_style():
        from .expose import Expose
        return Expose.application.config.ui.forms_style or FormStyle

    def __init__(self, fields={}, **kwargs):
        #: get fields from kwargs
        for name, parameter in iteritems(kwargs):
            if isinstance(parameter, Field):
                fields[name] = parameter
        for name in iterkeys(fields):
            if name in kwargs:
                del kwargs[name]
        #: order fields correctly
        sorted_fields = []
        for name, field in iteritems(fields):
            sorted_fields.append((name, field))
        sorted_fields.sort(key=lambda x: x[1]._inst_count_)
        #: init fields
        self.fields = []
        self.writable_fields = []
        for name, obj in sorted_fields:
            field_obj = obj._make_field(name)
            self.fields.append(field_obj)
            if field_obj.writable:
                self.writable_fields.append(field_obj)
        self._preprocess_(**kwargs)

    def _preprocess_(self, **kwargs):
        #: process attributes
        self.attributes = {}
        self.attributes.update(kwargs)
        for key, val in Form.default_attrs.items():
            self.attributes[key] = self.attributes.get(key, val)
        self.attributes['formstyle'] = self.attributes.get(
            'formstyle', self._get_default_style())
        #: init the form
        self.errors = sdict()
        self.params = sdict()
        self.input_params = None
        self.processed = False
        self.accepted = False
        self.formkey = "undef"
        #: move some attributes to self, just because it's handy
        self.keepvalues = self.attributes['keepvalues']
        self.onvalidation = self.attributes['onvalidation']
        del self.attributes['keepvalues']
        del self.attributes['onvalidation']
        #: verify formstyle consistence
        if not issubclass(self.attributes['formstyle'], FormStyle):
            raise RuntimeError(
                '%s is an invalid weppy form style' %
                self.attributes['formstyle'].__name__)
        #: process the form
        self._process()

    @property
    def csrf(self):
        _csrf = self.attributes["csrf"]
        return _csrf is True or (_csrf == 'auto' and
                                 self.attributes['_method'] == 'POST')

    def _load_csrf(self):
        if not self.csrf:
            return
        if not hasattr(current, "session"):
            raise RuntimeError("You need sessions to use csrf in forms.")
        session._csrf = session._csrf or CSRFStorage()

    @property
    def _submitted(self):
        if self.csrf:
            return self.input_params._csrf_token in session._csrf
        return self.input_params._csrf_token is 'undef'

    def _get_input_val(self, field):
        if field.type == 'boolean':
            v = self.input_params.get(field.name, False)
            if v is not False:
                v = True
        else:
            v = self.input_params.get(field.name)
        return v

    def _process(self):
        self._load_csrf()
        method = self.attributes['_method']
        # get appropriate input variables
        if method is "POST":
            self.input_params = sdict(request.body_params)
        else:
            self.input_params = sdict(request.query_params)
        # run processing if needed
        if self._submitted:
            self.processed = True
            # validate input
            for field in self.writable_fields:
                value = self._get_input_val(field)
                value, error = field.validate(value)
                if error:
                    self.errors[field.name] = error
                else:
                    self.params[field.name] = value
            # custom validation
            if not self.errors and callable(self.onvalidation):
                self.onvalidation(self)
            # end of validation
            if not self.errors:
                self.accepted = True
                if self.csrf:
                    del session._csrf[self.input_params._csrf_token]
        # CSRF protection logic
        if self.csrf and not self.accepted:
            self.formkey = session._csrf.gen_token()
        # reset default values in form
        if not self.processed or (self.accepted and not self.keepvalues):
            for field in self.fields:
                default_value = field.default() if callable(field.default) \
                    else field.default
                self.input_params[field.name] = default_value

    def _render(self):
        styler = self.attributes['formstyle'](self.attributes)
        styler.on_start()
        for field in self.fields:
            value = self.input_params.get(field.name)
            error = self.errors.get(field.name)
            styler._proc_element(field, value, error)
        styler.add_buttons()
        styler._add_formkey(self.formkey)
        for key, value in iteritems(self.attributes.get('hidden', {})):
            styler._add_hidden(key, value)
        return styler.render()

    @cachedprop
    def custom(self):
        # init
        custom = sdict()
        custom.dspval = sdict()
        custom.inpval = sdict()
        custom.label = sdict()
        custom.comment = sdict()
        custom.widget = sdict()
        # load selected styler
        styler = self.attributes['formstyle'](self.attributes)
        styler.on_start()
        # load data
        for field in self.fields:
            value = self.input_params.get(field.name)
            custom.dspval[field.name] = self.input_params[field.name]
            custom.inpval[field.name] = self.input_params[field.name]
            custom.label[field.name] = field.label
            custom.comment[field.name] = field.comment
            widget, wfield = styler._get_widget(field, value)
            if not wfield:
                styler.style_widget(widget)
            custom.widget[field.name] = widget
        # add submit
        custom.submit = tag.input(
            _type="submit", value=self.attributes['submit'])
        # provides begin attribute
        begin = '<form %s>' % self._build_html_attributes()
        custom.begin = asis(begin)
        # add hidden stuffs to get weppy process working
        hidden = cat()
        hidden.append(
            tag.input(
                _name='_csrf_token', _type='hidden', _value=self.formkey))
        for key, value in iteritems(self.attributes.get('hidden', {})):
            hidden.append(
                tag.input(_name=key, _type='hidden', _value=value))
        # provides end attribute
        end = '%s</form>' % hidden.to_html()
        custom.end = asis(end)
        return custom

    def to_html(self):
        return self._render().to_html()


class DALForm(Form):
    def __init__(self, table, record=None, record_id=None, fields=None,
                 exclude_fields=[], **attributes):
        self.table = table
        self.record = record or table(record_id)
        #: build fields for form
        self.fields = []
        self.writable_fields = []
        if fields is not None:
            #: developer has selected specific fields
            if not isinstance(fields, dict):
                fields = {'writable': fields, 'readable': fields}
            for field in table:
                if field.name not in fields['readable']:
                    continue
                self.fields.append(field)
                if field.name in fields['writable']:
                    self.writable_fields.append(field)
        else:
            #: use table fields
            for field in table:
                if field.name in exclude_fields:
                    continue
                if not field.readable:
                    continue
                if not self.record and not field.writable:
                    #: show readable fields only on update
                    continue
                self.fields.append(field)
                if field.writable:
                    self.writable_fields.append(field)
                # if field.type != 'id' and field.writable and \
                #         field.name not in exclude_fields:
                #     self.fields.append(field)
        #: use tablename for form id
        attributes['id_prefix'] = table._tablename + "_"
        #: finally init the form
        self._preprocess_(**attributes)

    def _process(self):
        #: send record id to validators if needed
        current._dbvalidation_record_id_ = None
        if self.record:
            current._dbvalidation_record_id_ = self.record.id
        #: load super `_process`
        Form._process(self)
        #: clear current and run additional operations for DAL
        del current._dbvalidation_record_id_
        if self.accepted:
            for field in self.writable_fields:
                #: handle uploads
                if field.type == 'upload':
                    f = self.params[field.name]
                    fd = field.name + "__del"
                    if f == b'' or f is None:
                        if self.input_params.get(fd, False):
                            self.params[field.name] = \
                                self.table[field.name].default or ''
                            # TODO: we want to physically delete file?
                        else:
                            if self.record and self.record[field.name]:
                                self.params[field.name] = \
                                    self.record[field.name]
                        continue
                    elif hasattr(f, 'file'):
                        source_file, original_filename = f.file, f.filename
                    else:
                        continue
                    newfilename = field.store(source_file, original_filename,
                                              field.uploadfolder)
                    if isinstance(field.uploadfield, str):
                        self.params[field.uploadfield] = source_file.read()
                    self.params[field.name] = newfilename
            #: add default values to hidden fields if needed
            if not self.record:
                fieldnames = [field.name for field in self.writable_fields]
                for field in self.table:
                    if field.name not in fieldnames and field.compute is None:
                        if field.default is not None:
                            def_val = field.default() \
                                if callable(field.default) else field.default
                            self.params[field.name] = def_val
            if self.record:
                self.record.update_record(**self.params)
            else:
                self.params.id = self.table.insert(**self.params)
        if not self.processed or (self.accepted and not self.keepvalues):
            for field in self.fields:
                if self.record:
                    self.input_params[field.name] = self.record[field.name]
                self.input_params[field.name] = field.formatter(
                    self.input_params[field.name])


class FormStyle(object):
    _stack = []
    parent = None

    @staticmethod
    def _field_options(field):
        if field.represent:
            def represent(value):
                if value and field.represent:
                    return field.represent(value)
                return value
        else:
            represent = lambda value: value

        validator = FormStyle._validation_woptions(field)
        option_items = [(k, represent(n)) for k, n in validator.options()]
        return option_items, validator.multiple

    @staticmethod
    def widget_string(attr, field, value, _class='string', _id=None):
        return tag.input(
            _type='text', _name=field.name,
            _value=value if value is not None else '',
            _class=_class, _id=_id or field.name
        )

    @staticmethod
    def widget_text(attr, field, value, _class='text', _id=None):
        return tag.textarea(value or '', _name=field.name,
                            _class=_class, _id=_id or field.name)

    @staticmethod
    def widget_int(attr, field, value, _class='int', _id=None):
        return FormStyle.widget_string(attr, field, value, _class, _id)

    @staticmethod
    def widget_bigint(attr, field, value, _class='int', _id=None):
        return FormStyle.widget_string(attr, field, value, _class, _id)

    @staticmethod
    def widget_float(attr, field, value, _class='float', _id=None):
        return FormStyle.widget_string(attr, field, value, _class, _id)

    @staticmethod
    def widget_date(attr, field, value, _class='date', _id=None):
        return FormStyle.widget_string(attr, field, value, _class, _id)

    @staticmethod
    def widget_time(attr, field, value, _class='time', _id=None):
        return FormStyle.widget_string(attr, field, value, _class, _id)

    @staticmethod
    def widget_datetime(attr, field, value, _class='datetime', _id=None):
        return FormStyle.widget_string(attr, field, value, _class, _id)

    @staticmethod
    def widget_password(attr, field, value, _class='password', _id=None):
        return tag.input(
            _type='password', _name=field.name, _value=value or '',
            _class=_class, _id=_id or field.name
        )

    @staticmethod
    def widget_bool(attr, field, value, _class='bool', _id=None):
        return tag.input(_type='checkbox', _name=field.name,
                         _checked='checked' if value else None,
                         _class=_class, _id=_id or field.name)

    @staticmethod
    def widget_select(attr, field, value, _class='', _id=None):
        def selected(k):
            return 'selected' if str(value) == str(k) else None

        options, multiple = FormStyle._field_options(field)
        if multiple:
            return FormStyle.widget_multiple(attr, field, value, options,
                                             _class=_class, _id=_id)
        option_items = [tag.option(n, _value=k, _selected=selected(k))
                        for k, n in options]
        return tag.select(*option_items, _name=field.name, _class=_class,
                          _id=_id or field.name)

    @staticmethod
    def widget_multiple(attr, field, values, options, _class='', _id=None):
        def selected(k):
            return 'selected' if str(k) in [str(v) for v in values] else None

        values = values or []
        option_items = [tag.option(n, _value=k, _selected=selected(k))
                        for k, n in options]
        return tag.select(*option_items, _name=field.name, _class=_class,
                          multiple='multiple', _id=_id or field.name)

    #: TO-DO
    #@staticmethod
    #def widget_list(attr, field, value, _class='', _id=None):
    #    return ""

    @staticmethod
    def widget_upload(attr, field, value, _class='upload',
                      _id=None):
        def is_image(value):
            extension = value.split('.')[-1].lower()
            if extension in ['gif', 'png', 'jpg', 'jpeg', 'bmp']:
                return True
            return False
        elements = []
        _value = value or ''
        download_url = attr.get('upload')
        inp = tag.input(_type='file', _name=field.name, _class=_class, _id=_id)
        elements.append(inp)
        if _value and download_url:
            if callable(download_url):
                url = download_url(value)
            else:
                url = download_url + '/' + value
            if is_image(_value):
                elements.append(tag.img(_src=url, _width='120px',
                                        _class='upload_img'))
            else:
                elements.append(tag.div(tag.a(_value, _href=url)))
            requires = field.requires or []
            # delete checkbox
            from .validators import isEmptyOr
            if not requires or (isinstance(v, isEmptyOr) for v in requires):
                elements.append(tag.div(
                    tag.input(_type='checkbox', _class='checkbox',
                              _id=_id + '__del', _name=field.name + '__del',
                              _style="display: inline;"),
                    tag.label('delete', _for=_id + '__del',
                              _style="margin: 4px"),
                    _style="white-space: nowrap;"))
        return tag.div(*elements, _class='upload_wrap')

    @staticmethod
    def widget_json(attr, field, value, _id=None):
        return FormStyle.widget_text(attr, field, value, _id=_id or field.name)

    def __init__(self, attributes):
        self.attr = attributes

    @staticmethod
    def _validation_woptions(field):
        ftype = field._type.split(":")[0]
        if ftype != "bool" and field.requires:
            for v in field.requires:
                if hasattr(v, 'options'):
                    return v
        return None

    #: returns the widget for the field and a boolean (True if widget is
    #  defined by user, False if it comes from styler default ones)
    def _get_widget(self, field, value):
        if field.widget:
            return field.widget(field, value), True
        widget_id = self.attr["id_prefix"] + field.name
        wtype = field._type.split(":")[0]
        if self._validation_woptions(field) is not None:
            wtype = 'select'
        elif wtype.startswith('reference'):
            wtype = 'int'
        try:
            widget = getattr(self, "widget_" + wtype)(
                self.attr, field, value, _id=widget_id)
            if not field.writable:
                self._disable_widget(widget)
            return widget, False
        except AttributeError:
            raise RuntimeError(
                "Missing form widget for field %s of type %s" %
                (field.name, wtype)
            )

    def _disable_widget(self, widget):
        widget.attributes['_disabled'] = 'disabled'

    def _proc_element(self, field, value, error):
        widget, wfield = self._get_widget(field, value)
        self._stack.append(sdict(widget=widget, _wffield=wfield))
        self._add_element(field.label, field.comment, error)
        self._stack.pop(-1)

    def _add_element(self, label, comment, error):
        # style only widgets not defined by user
        if not self.element._wffield:
            self.style_widget(self.element.widget)
        self.element.label = self.create_label(label)
        if comment:
            self.element.comment = self.create_comment(comment)
        if error:
            self.element.error = self.create_error(error)
        self.add_widget(self.element.widget)

    def _add_hidden(self, key, value):
        self.parent.append(tag.input(_name=key, _type='hidden', _value=value))

    def _add_formkey(self, key):
        self.parent.append(tag.input(_name='_csrf_token', _type='hidden',
                                     _value=key))

    @property
    def element(self):
        return self._stack[-1] if self._stack else None

    def on_start(self):
        self.parent = cat()

    def style_widget(self, widget):
        pass

    def create_label(self, label):
        wid = self.element.widget['_id']
        return tag.label(label, _for=wid, _class='wpp_label')

    def create_error(self, error):
        return tag.div(error, _class='wpp_error')

    def create_comment(self, comment):
        return tag.p(comment, _class='wpp_help')

    def add_widget(self, widget):
        wrapper = tag.div(widget)
        if self.element.error:
            wrapper.append(self.element.error)
        if self.element.comment:
            wrapper.append(self.element.comment)
        self.parent.append(tag.div(self.element.label, wrapper))

    def add_buttons(self):
        submit = tag.input(_type='submit', _value=self.attr['submit'])
        self.parent.append(tag.div(submit))

    def render(self):
        return tag.form(self.parent, **self.attr)



# -*- coding: utf-8 -*-
"""
    weppy._compat
    -------------

    Some py2/py3 compatibility support.

    :copyright: (c) 2014-2016 by Giovanni Barillari
    :license: BSD, see LICENSE for more details.
"""

import sys
import hashlib

PY2 = sys.version_info[0] == 2
_identity = lambda x: x

if PY2:
    from Cookie import SimpleCookie
    from cStringIO import StringIO
    import cPickle as pickle
    import copy_reg as copyreg
    iterkeys = lambda d: d.iterkeys()
    itervalues = lambda d: d.itervalues()
    iteritems = lambda d: d.iteritems()

    text_type = unicode
    integer_types = (int, long)
    string_types = (str, unicode)
    basestring = basestring

    reduce = reduce
    xrange = xrange
    hashlib_md5 = hashlib.md5
    hashlib_sha1 = hashlib.sha1

    def implements_iterator(cls):
        cls.next = cls.__next__
        del cls.__next__
        return cls

    def implements_bool(cls):
        cls.__nonzero__ = cls.__bool__
        del cls.__bool__
        return cls

    def implements_to_string(cls):
        cls.__unicode__ = cls.__str__
        cls.__str__ = lambda x: x.__unicode__().encode('utf-8')
        return cls

    exec('def reraise(tp, value, tb=None):\n raise tp, value, tb')

    def to_bytes(obj, charset='utf8', errors='strict'):
        if obj is None:
            return None
        if isinstance(obj, (bytes, bytearray, buffer)):
            return bytes(obj)
        if isinstance(obj, unicode):
            return obj.encode(charset, errors)
        raise TypeError('Expected bytes')

    def to_native(obj, charset='utf8', errors='strict'):
        if obj is None or isinstance(obj, str):
            return obj
        return obj.encode(charset, errors)

else:
    from http.cookies import SimpleCookie
    from io import StringIO
    import pickle
    import copyreg
    iterkeys = lambda d: iter(d.keys())
    itervalues = lambda d: iter(d.values())
    iteritems = lambda d: iter(d.items())

    text_type = str
    integer_types = (int, )
    string_types = (str,)
    basestring = str

    from functools import reduce
    xrange = range
    hashlib_md5 = lambda s: hashlib.md5(bytes(s, 'utf8'))
    hashlib_sha1 = lambda s: hashlib.sha1(bytes(s, 'utf8'))

    implements_iterator = _identity
    implements_bool = _identity
    implements_to_string = _identity

    def reraise(tp, value, tb=None):
        if value.__traceback__ is not tb:
            raise value.with_traceback(tb)
        raise value

    def to_bytes(obj, charset='utf8', errors='strict'):
        if obj is None:
            return None
        if isinstance(obj, (bytes, bytearray, memoryview)):
            return bytes(obj)
        if isinstance(obj, str):
            return obj.encode(charset, errors)
        raise TypeError('Expected bytes')

    def to_native(obj, charset='utf8', errors='strict'):
        if obj is None or isinstance(obj, str):
            return obj
        return obj.decode(charset, errors)


def with_metaclass(meta, *bases):
    """Create a base class with a metaclass."""
    # This requires a bit of explanation: the basic idea is to make a
    # dummy metaclass for one level of class instantiation that replaces
    # itself with the actual metaclass.
    class metaclass(type):
        def __new__(cls, name, this_bases, d):
            return meta(name, bases, d)
    return type.__new__(metaclass, 'temporary_class', (), {})


def to_unicode(obj, charset='utf8', errors='strict'):
    if obj is None:
        return None
    if not isinstance(obj, bytes):
        return text_type(obj)
    return obj.decode(charset, errors)
